import os
import sys
import numpy as np
from math import sqrt
from random import randint
from typing import List, Iterable, get_args

from . import config
from . import palette
from .colormath.color_conversions import convert_color
from .colormath.color_diff import *
from .colormath.color_objects import sRGBColor, LabColor
from .color_class import ColorBase, HCLab, ColorLike
from .color_format import ColorFormat
from .gradient import Grad

# Patch asscalar
np.asscalar = lambda a: a.item()

__all__ = [
    "swatch",
    "show",
    "simplified_dist",
    "color_dist",
    "random_color",
    "color_str",
    "hue_sort_key",
    "hue_sorted",
    "grayscale",
    "inverse",
    "blend"
]


def blend(color1, color2, perc=0.5, grad_class=Grad, **kwargs):
    return grad_class([color1, color2], **kwargs).perc(perc)


def grayscale(obj):
    """Returns a grayscale representation of a colorir object"""
    if isinstance(obj, list):
        return [color.grayscale() for color in obj]
    # Assume single colorlike
    elif isinstance(obj, get_args(ColorLike)):
        obj = config.DEFAULT_COLOR_FORMAT.format(obj)
    return obj.grayscale()


def inverse(obj):
    """Returns an RGB-inverted representation of a colorir object"""
    if isinstance(obj, list):
        return [~color for color in obj]
    # Assume single colorlike
    elif isinstance(obj, get_args(ColorLike)):
        obj = config.DEFAULT_COLOR_FORMAT.format(obj)
    return ~obj


def swatch(obj,
           colored_text=True,
           width=3,
           height=1,
           tabular=True,
           file=sys.stdout):
    """Swatches a colorir object in the terminal.

    Each swatch consists of a rectangle of a color followed by its value (and name if known).

    Args:
        obj: What will be represented in the terminal. Is either a single color, a list of colors,
            a :class:`~colorir.palette.Palette`, a :class:`~colorir.palette.StackPalette` or a
            subclass of :class:`~colorir.gradient.Grad`.
        colored_text: Whether the text that follows the colored rectangles should also be colored.
        width: The width (in space characters) of the colored rectangles.
        height: The height (in number of lines) of the colored rectangles.
        tabular: Whether the colored rectangle, color name and color value should be printed each
            in its separate column. Only used if `obj` is a :class:`~colorir.palette.Palette`.
        file: If ``None`` returns the swatch as a string. Otherwise, this argument must be a file object and
            is passed to `print`.
    """
    longest_name = None
    if isinstance(obj, Grad):
        # Seven steps for each color transition = 7 * (n - 1) - (n - 2)
        obj = obj.n_colors(6 * len(obj.colors) - 5, include_ends=True)
    elif isinstance(obj, palette.Palette):
        longest_name = max([len(name) for name in obj.color_names])
    # Assume single ColorLike
    elif not isinstance(obj, (list, palette.StackPalette)):
        obj = [obj]
    # Needed to make Windows understand "\33" (https://stackoverflow.com/questions/12492810/python-
    # how-can-i-make-the-ansi-escape-codes-to-work-also-in-windows)
    os.system("")
    ret_str = ""
    for i, c_val in enumerate(obj):
        rect_str = color_str(" " * width, bg_color=c_val)
        val_str = f" {c_val}"
        if longest_name is not None:
            name = obj.color_names[i]
            spacing = tabular * " " * (longest_name - len(name))
            val_str = ' ' + name + spacing + val_str
        if colored_text:
            val_str = color_str(val_str, fg_color=c_val)
        ret_str += rect_str + val_str + '\n'
        for _ in range(height - 1):
            ret_str += rect_str + '\n'
    ret_str = ret_str.strip("\n")
    if file is None:
        return ret_str
    print(ret_str, file=file)


def show(obj,
         width=None,
         height=None,
         interactive=None,
         inline=True):
    """Shows a colorir object inlined on a Jupyter notebook or using tkinter.

    Args:
        inline: If running on a Jupyter notebook, whether to display the object inline.
            Requires Pillow to be installed.
        obj:
        width:
        height:
        interactive: Whether clicking of the screen will copy the color value
            to the clipboard and swatch the color on the terminal. Setting this
            parameter to ``False`` results is a dramatic speed-up. By default,
            is set to ``True`` for small objects (less than 100 colors) and
            ``False`` otherwise. Only has an effect if not running on a
            Jupyter notebook with `inline` == ``True``.
    """
    if inline:
        try:
            from IPython.display import display
            # For some reason this function is not in globals()... so we gotta try it
            if get_ipython().__class__.__name__ == "ZMQInteractiveShell":
                display(make_image(obj, width, height))
                return
        except (ImportError, NameError):
            pass
        show_tkinter(obj, width, height, interactive)


def show_tkinter(obj, width, height, interactive):
    import tkinter as tk
    from . import tkinter_utils as tku

    win = tk.Tk()
    win.resizable(False, False)

    max_width = win.winfo_screenwidth() * 0.8
    max_height = win.winfo_screenheight() * 0.8

    if isinstance(obj, Grad):
        if width is None:
            width = min(max_width, max(0.4 * max_width, 50 * len(obj.colors)))
        if height is None:
            height = 0.1 * max_height
        width, height = int(width), int(height)

        # Each button is two pixels wide to make the widget lighter
        colors = obj.n_colors(round(width / 2))
        color_names = None
    else:
        if isinstance(obj, palette.Palette):
            colors = obj.colors
            color_names = obj.color_names
        elif isinstance(obj, (palette.StackPalette, list)):
            colors = list(obj)
            color_names = None
        else:
            colors = [config.DEFAULT_COLOR_FORMAT.format(obj)]
            color_names = None

        if height is None:
            height = 0.1 * max_height
        if width is None:
            width = min(max_width, len(colors) * height)
        width, height = int(width), int(height)

    if interactive is None:
        interactive = True if len(colors) < 100 else False
    if interactive:
        wgt = tku.PaletteWidget(
            win,
            colors,
            color_names=color_names,
            width=width,
            height=height
        )
    else:
        wgt = tku.StaticPaletteWidget(
            win,
            colors,
            width=width,
            height=height
        )
    wgt.pack()
    win.mainloop()


def make_image(obj, width, height):
    from PIL import Image

    if height is None:
        height = 64
    if isinstance(obj, Grad):
        if width is None:
            width = 512
        colors = obj.n_colors(width)
    else:
        if not isinstance(obj, (palette.PaletteBase, list)):
            obj = [config.DEFAULT_COLOR_FORMAT.format(obj)]
        colors = list(obj)
        if width is None:
            width = height * len(colors)
    img = np.array([config.DEFAULT_COLOR_FORMAT.format(color)._rgba for color in colors], dtype="uint8")
    # Add height dim
    img = np.reshape(img, (1, img.shape[0], 4))
    img = img.repeat(height, axis=0).repeat(width // len(colors), axis=1)
    return Image.fromarray(img, "RGBA")


def hue_sort_key(hue_classes=None,
                 gray_thresh=12.0,
                 gray_start=True,
                 alt_lum=False,
                 invert_lum=False):
    """Returns a function that can be used as a key for python's 'sort' and 'sorted' in order to
    sort color-like objects by their hue component.

    Examples:
        >>> sorted(["0000ff", "ff0000",  "000000", "00ff00", "ffffff"], key=hue_sort_key())
        ['000000', 'ffffff', 'ff0000', '00ff00', '0000ff']

    Args:
        hue_classes: Number hue categories. Inside each hue category colors will be sorted by
            luminance rather than hue. When ``None``, colors are sorted only according to
            their hue.
        gray_thresh: Chroma threshold bellow which a color will be considered a shade of gray.
        gray_start: Whether the colors considered shades of gray will be grouped at the start or
            end of the sorted iterable.
        alt_lum: Whether to alternate luminance values with each hue class transition. Only has an
            effect if `hue_classes` > 1.
        invert_lum: By default, sorting within hue classes happen from darker to lighter colors.
            This parameter allows inverting this patern, thus starting with light colors and going
            towards darker tones. Only has an effect if `hue_classes` > 1.

    Notes:
        The refered hue and chroma components are those of :class:`~colorir.color_class.HCLab`.
    """
    color_format = config.DEFAULT_COLOR_FORMAT
    gray_hue = -1 if gray_start else hue_classes + 1

    def sort_key(color):
        interpreted = color_format.format(color)
        h, c, l = HCLab._from_rgba(interpreted._rgba, max_h=1)
        if c < gray_thresh:
            h = gray_hue
            if not gray_start and alt_lum and hue_classes % 2 == 1:
                l *= -1
        elif hue_classes is not None:
            h = int(h * hue_classes)
            if alt_lum and h % 2 == (not gray_start):
                l *= -1

        if hue_classes is None:
            return h
        if invert_lum:
            l *= -1
        return h, l

    return sort_key


def hue_sorted(colors: Iterable[ColorLike], **kwargs) -> List[ColorLike]:
    """Sort colors by their hue values. See :func:`~colorir.utils.hue_sort_key` for the arguments docs.

    Returns:
        A list of sorted colors.
    """
    key = hue_sort_key(**kwargs)
    return sorted(colors, key=key)


def simplified_dist(color1: ColorLike,
                    color2: ColorLike):
    """Calculates the perceived distance between two colors.

    There are many methods to approach the similarity of colors mathematically. The
    algorithm implemented in this function [#]_ tries to provide balance between efficiency and
    accuracy by using a weighted euclidian distance formula in the RGB color space.

    References:
        .. [#] Adapted from "Colour metric" by Thiadmer Riemersma. Available on
            https://www.compuphase.com/cmetric.htm.

    Args:
        color1: First color point.
        color2: Second color point.
    """
    color_format = config.DEFAULT_COLOR_FORMAT
    # We only need the '._rgba's, so no need to convert if already ColorBase
    if not isinstance(color1, ColorBase):
        color1 = color_format.format(color1)
    if not isinstance(color2, ColorBase):
        color2 = color_format.format(color2)
    rgba1, rgba2 = color1._rgba, color2._rgba
    avg_r = (rgba1[0] + rgba2[0]) / (2 * 255)
    d_r = rgba1[0] - rgba2[0]
    d_g = rgba1[1] - rgba2[1]
    d_b = rgba1[2] - rgba2[2]
    return sqrt((2 + avg_r) * d_r ** 2
                + 4 * d_g ** 2
                + (3 - avg_r) * d_b ** 2)


# TODO doc (mention 2000 not working properly in colormath and kwargs for delta-e funcs)
def color_dist(color1: ColorLike, color2: ColorLike, method="CIE76"):
    if method == "simplified":
        return simplified_dist(color1, color2)

    color_format = config.DEFAULT_COLOR_FORMAT
    if not isinstance(color1, ColorBase):
        color1 = color_format.format(color1)
    if not isinstance(color2, ColorBase):
        color2 = color_format.format(color2)
    color1 = sRGBColor(*color1._rgba[:3], is_upscaled=True)
    color2 = sRGBColor(*color2._rgba[:3], is_upscaled=True)
    color1 = convert_color(color1, LabColor, target_illuminant="d65")
    color2 = convert_color(color2, LabColor, target_illuminant="d65")

    if method == "CIE76":
        return delta_e_cie1976(color1, color2)
    if method == "CIE94":
        return delta_e_cie1994(color1, color2)
    if method == "CIE2000":
        return delta_e_cie2000(color1, color2)
    if method == "CMC":
        return delta_e_cmc(color1, color2)

    raise ValueError("invalid 'method' parameter")


def random_color(random_a=False,
                 color_format: ColorFormat = None):
    """Generates a new random color.

    Args:
        random_a: Whether to randomize the alpha attribute as well or just make it 1.
        color_format: Specifies the format of the output color. Defaults to
            :data:`config.DEFAULT_COLOR_FORMAT <colorir.config.DEFAULT_COLOR_FORMAT>`.

    Examples:
        >>> random_color()  # doctest: +SKIP
        Hex('#304fcc')
    """
    if color_format is None:
        color_format = config.DEFAULT_COLOR_FORMAT
    if random_a:
        a = randint(0, 255)
    else:
        a = 255
    return color_format._from_rgba((randint(0, 255),
                                    randint(0, 255),
                                    randint(0, 255),
                                    a))


def color_str(string: str,
              fg_color: ColorLike = None,
              bg_color: ColorLike = None) -> str:
    """Returns an ANSI-escaped [#]_ colored representation of `string`.

    Examples:
        >>> print(color_str("It's Christmas!", "#ff0000", "#00ff00"))  # doctest: +SKIP

    Args:
        string: String that will be colored.
        fg_color: Color to be applied to the foreground of the text.
        bg_color: Color to be applied to the background of the text.

    References:
        .. [#] Wikipedia at https://en.wikipedia.org/wiki/ANSI_escape_code#Colors.
    """
    if fg_color is not None:
        rgba = config.DEFAULT_COLOR_FORMAT.format(fg_color)._rgba
        string = f"\033[38;2;{rgba[0]};{rgba[1]};{rgba[2]}m" + string + "\33[0m"
    if bg_color is not None:
        rgba = config.DEFAULT_COLOR_FORMAT.format(bg_color)._rgba
        string = f"\033[48;2;{rgba[0]};{rgba[1]};{rgba[2]}m" + string + "\33[0m"
    return string


# https://entropymine.com/imageworsener/srgbformula/
def _to_linear_rgb(rgba):
    """rgba must be in range 0-1"""
    srgb = np.array(rgba[:-1])
    lrgb = np.where(srgb <= 0.04045, srgb / 12.92, ((srgb + 0.055) / 1.055) ** 2.4)
    return np.pad(lrgb, [(0, 1)], constant_values=rgba[-1])


def _to_srgb(rgba):
    """rgba must be in range 0-1"""
    lrgb = np.array(rgba[:-1])
    srgb = np.where(lrgb <= 0.0031308, lrgb * 12.92, 1.055 * lrgb ** (1 / 2.4) - 0.055)
    return np.pad(srgb, [(0, 1)], constant_values=rgba[-1])


# -*- coding: utf-8 -*-
"""
The functions in this module are used for comparing two LabColor objects
using various Delta E formulas.
"""

import numpy

from . import color_diff_matrix


def _get_lab_color1_vector(color):
    """
    Converts an LabColor into a NumPy vector.

    :param LabColor color:
    :rtype: numpy.ndarray
    """
    if not color.__class__.__name__ == "LabColor":
        raise ValueError(
            "Delta E functions can only be used with two LabColor objects."
        )
    return numpy.array([color.lab_l, color.lab_a, color.lab_b])


def _get_lab_color2_matrix(color):
    """
    Converts an LabColor into a NumPy matrix.

    :param LabColor color:
    :rtype: numpy.ndarray
    """
    if not color.__class__.__name__ == "LabColor":
        raise ValueError(
            "Delta E functions can only be used with two LabColor objects."
        )
    return numpy.array([(color.lab_l, color.lab_a, color.lab_b)])


# noinspection PyPep8Naming
def delta_e_cie1976(color1, color2):
    """
    Calculates the Delta E (CIE1976) of two colors.
    """
    color1_vector = _get_lab_color1_vector(color1)
    color2_matrix = _get_lab_color2_matrix(color2)
    delta_e = color_diff_matrix.delta_e_cie1976(color1_vector, color2_matrix)[0]
    return delta_e.item()


# noinspection PyPep8Naming
def delta_e_cie1994(color1, color2, K_L=1, K_C=1, K_H=1, K_1=0.045, K_2=0.015):
    """
    Calculates the Delta E (CIE1994) of two colors.

    K_l:
      0.045 graphic arts
      0.048 textiles
    K_2:
      0.015 graphic arts
      0.014 textiles
    K_L:
      1 default
      2 textiles
    """
    color1_vector = _get_lab_color1_vector(color1)
    color2_matrix = _get_lab_color2_matrix(color2)
    delta_e = color_diff_matrix.delta_e_cie1994(
        color1_vector, color2_matrix, K_L=K_L, K_C=K_C, K_H=K_H, K_1=K_1, K_2=K_2
    )[0]
    return delta_e.item()


# noinspection PyPep8Naming
def delta_e_cie2000(color1, color2, Kl=1, Kc=1, Kh=1):
    """
    Calculates the Delta E (CIE2000) of two colors.
    """
    color1_vector = _get_lab_color1_vector(color1)
    color2_matrix = _get_lab_color2_matrix(color2)
    delta_e = color_diff_matrix.delta_e_cie2000(
        color1_vector, color2_matrix, Kl=Kl, Kc=Kc, Kh=Kh
    )[0]
    return delta_e.item()


# noinspection PyPep8Naming
def delta_e_cmc(color1, color2, pl=2, pc=1):
    """
    Calculates the Delta E (CMC) of two colors.

    CMC values
      Acceptability: pl=2, pc=1
      Perceptability: pl=1, pc=1
    """
    color1_vector = _get_lab_color1_vector(color1)
    color2_matrix = _get_lab_color2_matrix(color2)
    delta_e = color_diff_matrix.delta_e_cmc(color1_vector, color2_matrix, pl=pl, pc=pc)[
        0
    ]
    return delta_e.item()



"""Gradients between colors.

Examples:
    Get purple inbetween red and blue:

    >>> grad = Grad(["ff0000", "0000ff"])
    >>> grad(0.5)  # By default interpolates in CIELuv
    Hex('#be0090')

    Depending on your use-case, it may be useful to rather interpolate colors in a cilindrical
    color space, such as :class:`~colorir.color_class.HSV` or :class:`~colorir.color_class.HCLuv`.
    To do this, use :class:`PolarGrad`, which supports color lerping through polar coordinate
    color components:

    >>> p_grad = PolarGrad(["ff0000", "0000ff"])
    >>> p_grad(0.5)
    Hex('#f000cc')

    You can control how the hue is interpolated with the `hue_lerp` attribute:

    >>> p_grad.hue_lerp = "longest"  # Calculates the longest distance between the hue of red and blue
    >>> p_grad(0.5)  # Now we get green instead of purple
    Hex('#008800')

    Get 5 colors interspaced in the gradient:

    >>> grad.n_colors(5)
    [Hex('#ff0000'), Hex('#e00066'), Hex('#be0090'), Hex('#9400b9'), Hex('#0000ff')]

    You can use the 'domain' argument to change the range in which values are interpolated:

    >>> grad = Grad(["ff0000", "0000ff"], domain=[4, 8])
    >>> grad(6)
    Hex('#be0090')

    And the 'color_coords' argument to specify the position of the input colors in the gradient:

    >>> grad = Grad(["ff0000", "00ff00", "0000ff"], color_coords=[0, 0.75, 1])  # Green sits closer to blue than to red
    >>> grad(0.75)
    Hex('#00ff00')
"""

import numpy as np
from copy import copy, deepcopy
from typing import Iterable, Type, List
from . import config, utils
from .color_class import RGB, CIELuv, ColorBase, HCLuv, ColorPolarBase, Hex, ColorLike
from .color_format import MATPLOTLIB_COLOR_FORMAT

__all__ = [
    "Grad",
    "PolarGrad",
    "RGBGrad",
    "RGBLinearGrad"
]


class Grad:
    """Linear interpolation gradient.

    Args:
        colors: Iterable of colors that compose the gradient.
        color_format: Color format specifying how to output colors. Defaults to
            :const:`config.DEFAULT_COLOR_FORMAT <colorir.config.DEFAULT_COLOR_FORMAT>`.
        color_sys: Color system in which the colors will be interpolated.
        domain: Range (inclusive on both ends) from which colors can be interpolated.
        color_coords: A list of the coordinates where each color in 'colors' is located in the gradient.
            If left empty, interspaced points will be automatically generated. For example,
            the default for a list of three colors with domain == [0, 1] is [0, 0.5, 1].
            Must have the same length as the `colors` argument.
        discrete: If ``True``, does not interpolate the colors and instead return the
            closest color to the input value (according to `color_coords`). Useful to plot
            categorical data.
    """

    def __init__(self,
                 colors: Iterable[ColorLike],
                 color_sys: Type[ColorBase] = CIELuv,
                 color_format=None,
                 color_coords: List[float] = None,
                 domain=(0, 1),
                 discrete=False):
        if color_sys == Hex:
            raise ValueError("only tuple-based color systems are supported")
        if domain[0] >= domain[1]:
            raise ValueError("'domain' must follow the format [lower_bound, upper_bound]")
        if color_coords is None:
            color_coords = np.linspace(domain[0], domain[1], len(colors))
        if len(color_coords) != len(colors):
            raise ValueError("'color_coords' must have same length as 'colors'")
        if color_format is None:
            color_format = config.DEFAULT_COLOR_FORMAT

        self.color_sys = color_sys
        self.color_format = color_format
        self.domain = list(domain)
        self.color_coords = np.array(color_coords)
        self.discrete = discrete
        self.colors = colors

    @property
    def colors(self):
        return list(self._colors)

    @colors.setter
    def colors(self, value):
        self._colors = [self.color_format.format(color) for color in value]
        self._update_conv_colors()

    @property
    def color_edges(self):
        edges = []
        for c1, c2 in zip(self.color_coords, self.color_coords[1:]):
            edges.append((c1 + c2) / 2)
        return edges

    def __str__(self):
        return f"{self.__class__.__name__}({self._colors})"

    def __repr__(self):
        if config.REPR_STYLE in ["traditional", "inherit"]:
            return str(self)
        return utils.swatch(self, file=None)

    def __call__(self, x, restrict_domain=False):
        return self.at(x, restrict_domain=restrict_domain)

    # noinspection PyDefaultArgument
    def __deepcopy__(self, memodict={}):
        obj = copy(self)
        obj._colors = list(obj._colors)
        obj._conv_colors = list(obj._conv_colors)
        obj.domain = list(obj.domain)
        obj.color_coords = np.array(obj.color_coords)
        return obj

    def __invert__(self):
        obj = deepcopy(self)
        obj.colors = [~color for color in obj._colors]
        return obj

    def grayscale(self):
        obj = deepcopy(self)
        obj.colors = [color.grayscale() for color in obj._colors]
        return obj

    def at(self, x, restrict_domain=False) -> ColorBase:
        if restrict_domain and (x < self.domain[0] or x > self.domain[1]):
            raise ValueError("'x' is out of the gradient domain")
        i = min(np.digitize(x, self.color_coords), len(self._colors) - 1)
        if i == 0:
            return self.color_format._from_rgba(self._conv_colors[0]._rgba)
        p = (x - self.color_coords[i - 1]) / (self.color_coords[i] - self.color_coords[i - 1])
        if p > 1:
            return self.color_format._from_rgba(self._conv_colors[-1]._rgba)
        if self.discrete:
            return self.color_format._from_rgba(self._conv_colors[i - 1 + round(p)]._rgba)
        new_color = self._linear_interp(
            self._conv_colors[i - 1],
            self._conv_colors[i],
            p
        )
        # Sometimes this reinterpretation makes colors seem to not be linearly-interpolated
        # HCLuv(287, 99, 31) -> RGB(131, 0, 185) -> HCLuv(286, 95, 35)
        # This is okay though
        return self.color_format._from_rgba(new_color._rgba)

    def perc(self, p: float, restrict_domain=False) -> ColorBase:
        """Returns the color placed in a given percentage of the gradient.

        Args:
            restrict_domain:
            p: Percentage of the gradient expressed in a range of 0-1 from which a color will be
                drawn.

        Examples:
            Get purple inbetween red and blue:

            >>> grad = Grad([RGB(1, 0, 0), RGB(0, 0, 1)])
            >>> grad.perc(0.5)
            Hex('#be0090')

            Get a very "reddish" purple:

            >>> grad.perc(0.2)
            Hex('#e6005c')
        """
        x = p * (self.domain[1] - self.domain[0]) + self.domain[0]
        return self.at(x, restrict_domain=restrict_domain)

    def n_colors(self, n: int, include_ends=True) -> List[ColorBase]:
        """Return `n` interspaced colors from the gradient.

        Args:
            n: Number of colors to retrieve.
            include_ends: Stops the colors at the extreme of the gradient from being sampled.
                This allows sampling a small number of colors (such as two) without having it
                return the same colors that were used to create the gradient in the first place.
        """
        if n < 1:
            raise ValueError("'n' must be a positive integer")
        if n == 1:
            ps = [0.5]
        elif include_ends:
            ps = np.linspace(0, 1, n)
        else:
            ps = np.linspace(0, 1, n + 2)[1:-1]
        return [self.perc(p) for p in ps]

    def to_cmap(self, name=None, n=256):
        """Converts this gradient into a matplotlib ListedColormap.

        Args:
            n: Number of discrete colors in the resulting color map.
            name: Passed down to color map constructor.
        """
        from matplotlib.colors import ListedColormap

        colors = [MATPLOTLIB_COLOR_FORMAT.format(color) for color in self.n_colors(n)]
        return ListedColormap(colors, N=n, name=name)

    def to_plotly_colorscale(self, n=256):
        """Make a color scale according to plotly's format definition.

        Args:
            n: Number of internal interpolations to perform before creating the color scale. Increases
                the accuracy of the color scale. This argument is not used if the gradient is discrete.
        """
        colorscale = []
        if not self.discrete:
            for p in np.linspace(0, 1, n):
                colorscale.append((p, self.perc(p).hex()))
            return colorscale
        scale_pos = [0] + self.color_edges + [1]
        for i, pos in enumerate(scale_pos[:-1]):
            color = self.colors[i].hex()
            colorscale.append((pos, color))
            colorscale.append((scale_pos[i + 1], color))
        return colorscale

    def _linear_interp(self, color1, color2, p: float) -> ColorBase:
        """Receives two colors in 'self.color_sys' format (expected to include_a) and returns a
        new color in the same format."""
        color1, color2 = np.array(color1), np.array(color2)
        return self.color_sys(
            *(color1 + (color2 - color1) * p)
        )

    def _update_conv_colors(self):
        self._conv_colors = []
        for color in self._colors:
            color = self.color_format.format(color)
            color = self.color_sys._from_rgba(color._rgba, include_a=True, round_to=-1)
            self._conv_colors.append(color)


class PolarGrad(Grad):
    """Similar to `Grad` but can calculate the shortest path for hue interpolation.

    Args:
        hue_lerp: Which method to apply to hue interpolation. Options are: ``None``, "increase",
            "decrease", "shortest" and "longest".
            ``None`` means that the hue will be treated as any other color component and interpolated normally.
            "increment" and "decrement" indicate that the hue can only be incremented or
            decremented respectively.
            "shortest" and "longest" make sure that the hue is interpolated in the shortest or longest
            paths between the colors. Default is "shortest".
        colors: Iterable of colors that compose the gradient.
        color_sys: Color system in which the colors will be interpolated.
    """
    def __init__(self,
                 colors,
                 color_sys=HCLuv,
                 hue_lerp="shortest",
                 **kwargs):
        if not issubclass(color_sys, ColorPolarBase):
            raise ValueError("'color_sys' must be a subclass of 'ColorPolarBase'")
        if hue_lerp not in [None, "increment", "decrement", "shortest", "longest"]:
            raise ValueError("'hue_lerp' must be None, 'increment', 'decrement', 'shortest' or 'longest'")

        super().__init__(colors=colors, color_sys=color_sys, **kwargs)
        self.hue_lerp = hue_lerp

    def _linear_interp(self, color1, color2, p: float) -> ColorBase:
        d = abs(color1.h - color2.h)
        if self.hue_lerp is None \
                or (self.hue_lerp == "shortest" and d <= color1.max_h / 2) \
                or (self.hue_lerp == "longest" and d > color1.max_h / 2):
            return super()._linear_interp(color1=color1, color2=color2, p=p)

        array1, array2 = np.array(color1), np.array(color2)
        if self.hue_lerp == "increment" or (self.hue_lerp != "decrement" and color1.h > color2.h):
            array1, array2 = array2, array1
            p = 1 - p
        array1[0] += color1.max_h
        new_color = array1 + (array2 - array1) * p
        new_color[0] %= color1.max_h

        return self.color_sys(*new_color)


class RGBGrad(Grad):
    """Linear interpolation gradient using the RGB color space.

    This class allows colors to be interpolated using linear RGB through the 'use_linear_rgb'
    parameter. Otherwise, ``RGBGrad(..., use_linear_rgb=False)`` is the same as
    ``Grad(..., color_sys=sRGB)``.

    Args:
        colors: Iterable of colors that compose the gradient. There can be more than two colors in
            this iterable.
        color_format: Color format specifying how to output colors. Defaults to
            :const:`config.DEFAULT_COLOR_FORMAT <colorir.config.DEFAULT_COLOR_FORMAT>`.
        use_linear_rgb: Whether to use linear RGB rather than sRGB to create the gradient.
            Defaults to ``True``.

    Notes:
        Often using linear RGB may result in a gradient that looks more natural to the human eye
        [#]_. THIS IS THE DEFAULT.

    References:
        .. [#] Ayke van Laethem at https://aykevl.nl/2019/12/colors
    """

    def __init__(self, colors: Iterable[ColorLike], use_linear_rgb=True, **kwargs):
        super().__init__(colors=colors, color_sys=RGB, **kwargs)
        self.use_linear_rgb = use_linear_rgb

    def _linear_interp(self, color1, color2, p: float) -> ColorBase:
        rgba_1 = color1._rgba
        rgba_2 = color2._rgba
        if self.use_linear_rgb:
            rgba_1 = utils._to_linear_rgb(rgba_1 / 255)
            rgba_2 = utils._to_linear_rgb(rgba_2 / 255)

        new_rgba = rgba_1 + (rgba_2 - rgba_1) * p
        if self.use_linear_rgb:
            new_rgba = utils._to_srgb(new_rgba) * 255
        return RGB._from_rgba(new_rgba)


# Alias
RGBLinearGrad = RGBGrad



# -*- coding: utf-8 -*-
"""
Formulas for density calculation.
"""

from math import log10
from .density_standards import (
    ANSI_STATUS_T_BLUE,
    ANSI_STATUS_T_GREEN,
    ANSI_STATUS_T_RED,
    VISUAL_DENSITY_THRESH,
    ISO_VISUAL,
)


def ansi_density(color, density_standard):
    """
    Calculates density for the given SpectralColor using the spectral weighting
    function provided. For example, ANSI_STATUS_T_RED. These may be found in
    :py:mod:`colormath.density_standards`.

    :param SpectralColor color: The SpectralColor object to calculate
        density for.
    :param numpy.ndarray density_standard: NumPy array of filter of choice
        from :py:mod:`colormath.density_standards`.
    :rtype: float
    :returns: The density value for the given color and density standard.
    """
    # Load the spec_XXXnm attributes into a Numpy array.
    sample = color.get_numpy_array()
    # Matrix multiplication
    intermediate = sample * density_standard

    # Sum the products.
    numerator = intermediate.sum()
    # This is the denominator in the density equation.
    sum_of_standard_wavelengths = density_standard.sum()

    # This is the top level of the density formula.
    return -1.0 * log10(numerator / sum_of_standard_wavelengths)


def auto_density(color):
    """
    Given a SpectralColor, automatically choose the correct ANSI T filter.
    Returns a tuple with a string representation of the filter the
    calculated density.

    :param SpectralColor color: The SpectralColor object to calculate
        density for.
    :rtype: float
    :returns: The density value, with the filter selected automatically.
    """
    blue_density = ansi_density(color, ANSI_STATUS_T_BLUE)
    green_density = ansi_density(color, ANSI_STATUS_T_GREEN)
    red_density = ansi_density(color, ANSI_STATUS_T_RED)

    densities = [blue_density, green_density, red_density]
    min_density = min(densities)
    max_density = max(densities)
    density_range = max_density - min_density

    # See comments in density_standards.py for VISUAL_DENSITY_THRESH to
    # understand what this is doing.
    if density_range <= VISUAL_DENSITY_THRESH:
        return ansi_density(color, ISO_VISUAL)
    elif blue_density > green_density and blue_density > red_density:
        return blue_density
    elif green_density > blue_density and green_density > red_density:
        return green_density
    else:
        return red_density



# -*- coding: utf-8 -*-
"""
This module contains exceptions for use throughout the L11 Colorlib.
"""


class ColorMathException(Exception):
    """
    Base exception for all colormath exceptions.
    """

    pass


class UndefinedConversionError(ColorMathException):
    """
    Raised when the user asks for a color space conversion that does not exist.
    """

    def __init__(self, cobj, cs_to):
        super(UndefinedConversionError, self).__init__(cobj, cs_to)
        self.message = "Conversion from %s to %s is not defined." % (cobj, cs_to)


class InvalidIlluminantError(ColorMathException):
    """
    Raised when an invalid illuminant is set on a ColorObj.
    """

    def __init__(self, illuminant):
        super(InvalidIlluminantError, self).__init__(illuminant)
        self.message = "Invalid illuminant specified: %s" % illuminant


class InvalidObserverError(ColorMathException):
    """
    Raised when an invalid observer is set on a ColorObj.
    """

    def __init__(self, cobj):
        super(InvalidObserverError, self).__init__(cobj)
        self.message = "Invalid observer angle specified: %s" % cobj.observer



import doctest
import unittest
import numpy as np

from colorir import *

config.REPR_STYLE = "traditional"


def load_tests(loader, tests, ignore):
    tests.addTests(doctest.DocTestSuite(color_class))
    return tests


class TestConversion(unittest.TestCase):
    rgba = np.array([50, 78, 5, 255])

    def test_rgb_conversion(self):
        color = RGB(50, 78, 5, max_rgb=255)
        self.assertEqual(
            color,
            RGB._from_rgba(self.rgba)
        )
        self.assertEqual(
            color,
            (50, 78, 5)
        )

    def test_hsl_conversion(self):
        color = HSL(0.23059, 0.87950, 0.16275, max_h=1)
        self.assertEqual(
            color,
            HSL._from_rgba(self.rgba)
        )
        self.assertEqual(
            color,
            (0.23059, 0.87950, 0.16275)
        )

    def test_hsv_conversion(self):
        color = HSV(0.23059, 0.93588, 0.30588, max_h=1)
        self.assertEqual(
            color,
            HSV._from_rgba(self.rgba)
        )
        self.assertEqual(
            color,
            (0.23059, 0.93588, 0.30588)
        )

    def test_cmy_conversion(self):
        color = CMY(0.80392, 0.69412, 0.98039)
        self.assertEqual(
            color,
            CMY._from_rgba(self.rgba)
        )
        self.assertEqual(
            color,
            (0.80392, 0.69412, 0.98039)
        )

    def test_cmyk_conversion(self):
        color = CMYK(0.35897, 0, 0.9359, 0.69412)
        self.assertEqual(
            color,
            CMYK._from_rgba(self.rgba),
            (0.35897, 0, 0.9359, 0.69412)
        )

    def test_cielab_conversion(self):
        color = CIELab(29.757, -22.344, 35.474)
        self.assertEqual(
            color,
            CIELab._from_rgba(self.rgba)
        )
        self.assertEqual(
            color,
            (29.757, -22.344, 35.474)
        )

    def test_cieluv_conversion(self):
        color = CIELuv(29.757, -13.267, 33.646)
        self.assertEqual(
            color,
            CIELuv._from_rgba(self.rgba)
        )
        self.assertEqual(
            color,
            (29.757, -13.267, 33.646)
        )

    def test_hcluv_conversion(self):
        color = HCLuv(111.514, 36.172, 29.757)
        self.assertEqual(
            color,
            HCLuv._from_rgba(self.rgba)
        )
        self.assertEqual(
            color,
            (111.514, 36.172, 29.757)
        )

    def test_hclab_conversion(self):
        color = HCLab(122.206, 41.926, 29.757)
        self.assertEqual(
            color,
            HCLab._from_rgba(self.rgba)
        )
        self.assertEqual(
            color,
            (122.206, 41.926, 29.757)
        )

    def test_hex_conversion(self):
        color = Hex("#324e05")
        self.assertEqual(
            color,
            Hex._from_rgba(self.rgba)
        )
        self.assertEqual(
            color,
            "#324e05"
        )


if __name__ == "__main__":
    unittest.main()


"""Common color spaces and systems.

This module contains different classes that represent a few of the most common color spaces [#]_.
Colors can be compared, converted to and passed as arguments to different classes.

Examples:
    Create some colors and compare them:

    >>> RGB(1, 0, 0) == HSL(0, 1 , 0.5)
    True

    Convert an RGB color to CMYK:

    >>> rgb_red = RGB(1, 0, 0)
    >>> rgb_red.cmyk()
    CMYK(0.0, 1.0, 1.0, 0.0)

    Color arithmetic makes it easy to change the properties of a color.
    The operations are performed element-wise in the format of the color used as the right operand,
    allowing specific color components to be changed in the color on the left of the operator.
    If the right operand is not a color, but rather a tuple, than the left color is not converted
    prior to the operation being performed.

    Add some CIE lightness to the color black (right operand is CIELab, meaning that the Hex color will be converted
    to CIELab before adding 50 to its lighness value and then converted back to Hex):

    >>> Hex("#000000") + CIELab(50, 0, 0)
    Hex('#777777')

    Make a red 50% less saturated by multiplying its saturation component by 0.5 (here no conversion is performed
    because the right operand is a tuple rather than a color class):

    >>> HSV(0.0, 1.0, 1.0) * (1, 0.5, 1)
    HSV(0.0, 0.5, 1.0)

References:
    .. [#] Wikipedia at https://en.wikipedia.org/wiki/Color_model.
"""
import abc
import colorsys
import operator
import warnings

import numpy as np
from typing import List, Union
from .colormath.color_objects import (
    LabColor,
    LuvColor,
    LCHuvColor,
    sRGBColor,
    LCHabColor,
    CMYColor,
    CMYKColor
)
from .colormath.color_conversions import convert_color

import colorir

__all__ = [
    "ColorLike",
    "ColorBase",
    "ColorTupleBase",
    "ColorPolarBase",
    "RGB",
    "sRGB",
    "HSV",
    "HSL",
    "CMY",
    "CMYK",
    "CIELuv",
    "CIELab",
    "HCLuv",
    "HCLab",
    "Hex"
]


class ColorBase(metaclass=abc.ABCMeta):
    """Base class from which all color classes inherit.

    Notes:
        This class is abstract and should not be instantiated.
    """
    _rgba: np.ndarray
    _format_params: List[str]

    # Factory method to be called when reading the palette files or reconstructing colors
    @classmethod
    @abc.abstractmethod
    def _from_rgba(cls, rgba, **kwargs):
        pass

    @property
    def format(self) -> "colorir.color_format.ColorFormat":
        """Returns a :class:`~colorir.color_format.ColorFormat` representing the format of this
        color object."""
        format_ = {param: getattr(self, param) for param in self._format_params}
        return colorir.color_format.ColorFormat(self.__class__, **format_)

    def __eq__(self, other):
        try:
            if not isinstance(other, ColorBase):
                other = colorir.config.DEFAULT_COLOR_FORMAT.format(other)
            return np.all(np.rint(self._rgba) == np.rint(other._rgba))
        except colorir.color_format.FormatError:
            return NotImplemented

    def __hash__(self):
        return hash(tuple(self._rgba))

    def __invert__(self):
        """Gets the inverse RGB of this color."""
        return self.format._from_rgba(np.append(255 - self._rgba[:-1], self._rgba[-1]))

    def __mod__(self, other):
        """Blends two colors at 50% using :func:`colorir.utils.blend()`."""
        return colorir.blend(self, other)

    def __add__(self, other):
        """Adds two colors or a color (as the left operand) and a tuple (as the right operand) together."""
        return self._arithm_func(other, operator.add)

    def __sub__(self, other):
        """Subtracts one color from another or a tuple (the right operand) from a color (the left operand)."""
        return self._arithm_func(other, operator.sub)

    def __mul__(self, other):
        """Multiplies two colors or a color (as the left operand) and a tuple (as the right operand) together."""
        return self._arithm_func(other, operator.mul)

    def __truediv__(self, other):
        """Divides two colors or a color (as the left operand) and a tuple (as the right operand) together."""
        return self._arithm_func(other, operator.truediv)

    def grayscale(self):
        """Converts this color to a grayscale representation in the same format using CIE
        lightness component."""
        return self.format.format(CIELuv(self.cieluv().l, 0, 0))

    def hex(self, **kwargs) -> "Hex":
        """Converts the current color to a hexadecimal representation.

        Args:
            **kwargs: Keyword arguments wrapped in this function will be passed on to the
                :class:`Hex` constructor.
        """
        return Hex._from_rgba(self._rgba, **kwargs)

    def rgb(self, **kwargs) -> "sRGB":
        """Converts the current color to an sRGB representation.

        Args:
            **kwargs: Keyword arguments wrapped in this function will be passed on to the
                :class:`sRGB` constructor.
        """
        return sRGB._from_rgba(self._rgba, **kwargs)

    def hsl(self, **kwargs) -> "HSL":
        """Converts the current color to an HSL representation.

        Args:
            **kwargs: Keyword arguments wrapped in this function will be passed on to the
                :class:`HSL` constructor.
        """
        return HSL._from_rgba(self._rgba, **kwargs)

    def hsv(self, **kwargs) -> "HSV":
        """Converts the current color to an HSV representation.

        Args:
            **kwargs: Keyword arguments wrapped in this function will be passed on to the
                :class:`HSV` constructor.
        """
        return HSV._from_rgba(self._rgba, **kwargs)

    def cmyk(self, **kwargs) -> "CMYK":
        """Converts the current color to a CMYK representation.

        Args:
            **kwargs: Keyword arguments wrapped in this function will be passed on to the
                :class:`CMYK` constructor.
        """
        return CMYK._from_rgba(self._rgba, **kwargs)

    def cmy(self, **kwargs) -> "CMY":
        """Converts the current color to a CMY representation.

        Args:
            **kwargs: Keyword arguments wrapped in this function will be passed on to the
                :class:`CMY` constructor.
        """
        return CMY._from_rgba(self._rgba, **kwargs)

    def cieluv(self, **kwargs) -> "CIELuv":
        """Converts the current color to a CIELuv representation.

        Args:
            **kwargs: Keyword arguments wrapped in this function will be passed on to the
                :class:`CIELuv` constructor.
        """
        return CIELuv._from_rgba(self._rgba, **kwargs)

    def cielab(self, **kwargs) -> "CIELab":
        """Converts the current color to a CIELab representation.

        Args:
            **kwargs: Keyword arguments wrapped in this function will be passed on to the
                :class:`CIELab` constructor.
        """
        return CIELab._from_rgba(self._rgba, **kwargs)

    def hcluv(self, **kwargs) -> "HCLuv":
        """Converts the current color to a HCLuv representation.

        Args:
            **kwargs: Keyword arguments wrapped in this function will be passed on to the
                :class:`HCLuv` constructor.
        """
        return HCLuv._from_rgba(self._rgba, **kwargs)

    def hclab(self, **kwargs) -> "HCLab":
        """Converts the current color to a HCLab representation.

        Args:
            **kwargs: Keyword arguments wrapped in this function will be passed on to the
                :class:`HCLab` constructor.
        """
        return HCLab._from_rgba(self._rgba, **kwargs)

    def _arithm_func(self, other, foo):
        if not isinstance(other, ColorTupleBase):
            raise ValueError("operations are only possible if one of the colors is tuple-based")
        vals = tuple(map(foo, other.format.format(self), other))
        return self.format.format(other.format.format(vals))


class ColorTupleBase(ColorBase, tuple, metaclass=abc.ABCMeta):
    """Base class from which all color classes that are represented by tuples inherit.

    Notes:
        This class is abstract and should not be instantiated.
    """

    @abc.abstractmethod
    def __new__(cls, specs, a, rgba, include_a, round_to):
        specs = list(specs)
        if round_to == 0:
            specs = [round(val) for val in specs]
            a = round(a)
        elif round_to > 0:
            specs = [round(val, round_to) for val in specs]
            a = round(a, round_to)
        if include_a:
            specs.append(a)
        obj = tuple.__new__(cls, specs)
        obj.a = a
        obj.include_a = include_a
        obj.round_to = round_to
        obj._rgba = np.rint(rgba).astype(int)
        obj._format_params = ["include_a", "round_to"]

        return obj

    # It would be dangerous to change str conversion as the target framework could call it
    # expecting (255, 0, 0) and get RGB(255, 0, 0)
    def __str__(self):
        return tuple.__repr__(self)

    def __repr__(self):
        if colorir.config.REPR_STYLE == "traditional":
            return f"{self.__class__.__name__}{tuple.__repr__(self)}"
        if colorir.config.REPR_STYLE == "inherit":
            return str(self)
        return colorir.utils.swatch(self, file=None)

    def __hash__(self):
        return ColorBase.__hash__(self)

    def __eq__(self, other):
        colorbase_eq = ColorBase.__eq__(self, other)
        # If other is ColorBase than we trust the result of eq
        if isinstance(other, ColorBase):
            return colorbase_eq
        # Otherwise we also try tuple.__eq__
        return any([colorbase_eq is True, tuple.__eq__(self, other) is True])

    def __add__(self, other):
        return self._tup_arithm_func(other, operator.add)

    def __sub__(self, other):
        return self._tup_arithm_func(other, operator.sub)

    def __mul__(self, other):
        return self._tup_arithm_func(other, operator.mul)

    def __truediv__(self, other):
        return self._tup_arithm_func(other, operator.truediv)

    # If right side of the operator is not a tuple-based color, we perform the operation element-wise
    def _tup_arithm_func(self, other, foo):
        if not isinstance(other, ColorBase):
            vals = tuple(map(foo, self, other))
            return self.format.format(vals)
        return ColorBase._arithm_func(self, other, foo)


class ColorPolarBase(ColorTupleBase, metaclass=abc.ABCMeta):
    """Mixin tag indicating that the color contains a polar hue component.

    Notes:
        This class is abstract and should not be instantiated.
    """
    h: float
    max_h: float


class RGB(ColorTupleBase):
    """Represents a color in the RGB color space [#]_.

    References:
        .. [#] Wikipedia at https://en.wikipedia.org/wiki/SRGB.

    Args:
        r: Red component of the color.
        g: Green component of the color.
        b: Blue component of the color.
        a: Opacity component of the color. Defaults to ``None``, which means it will be the same
            as the `max_rgba` parameter.
        max_rgb: What is the maximum value for the `r`, `g` and `b` components. Some common
            values for this parameter would be 255 or 1.
        max_a : What is the maximum value for the `a` component. Some common
            values for this parameter would be 100 or 1.
        include_a: Whether to include the opacity parameter `a` in the constructed tuple.
            Setting it to ``True`` may result in an object such as :code:`RGB(255, 255, 0,
            255)` instead of :code:`RGB(255, 255, 0)`, for example.
        round_to: Rounds the value of each color component to this many decimal places. Setting
            this parameter to 0 ensures that the components will be of type `int`. -1
            means that the components won't be rounded at all.
        linear: Whether the values are linear RGB or sRGB. It is strongly advised not to keep values as
            linear RGB, but it can be useful for quick conversions.
    """

    def __new__(cls,
                r: float,
                g: float,
                b: float,
                a: float = None,
                max_rgb=1,
                max_a=1,
                include_a=False,
                round_to=-1,
                linear=False):
        if a is None:
            a = max_a
        elif not 0 <= a <= max_a:
            raise ValueError("'a' must be greater than 0 and smaller than 'max_a'")
        if not all(0 <= spec <= max_rgb for spec in (r, g, b)):
            raise ValueError("'r', 'g' and 'b' must be greater than 0 and smaller than 'max_rgb'")

        rgba = np.array((r, g, b, a), dtype=float)
        rgba[:3] /= max_rgb
        rgba[-1] /= max_a
        if linear:
            rgba = colorir.utils._to_srgb(rgba)
        rgba *= 255

        obj = super().__new__(
            cls,
            (r, g, b),
            a,
            rgba,
            include_a=include_a,
            round_to=round_to
        )
        obj.r, obj.g, obj.b = obj[:3]
        obj.max_rgb = max_rgb
        obj.max_a = max_a
        obj.linear = linear
        obj._format_params += ["max_rgb", "max_a", "linear"]

        return obj

    @classmethod
    def _from_rgba(cls, rgba, max_rgb=1, max_a=1, include_a=False, round_to=-1, linear=False):
        rgb = rgba / 255
        if linear:
            rgb = colorir.utils._to_linear_rgb(rgb)
        rgb = rgb[:-1]
        rgb *= max_rgb

        obj = super().__new__(cls,
                              rgb,
                              rgba[-1] / 255 * max_a,
                              rgba,
                              include_a=include_a,
                              round_to=round_to)
        obj.r, obj.g, obj.b = obj[:3]
        obj.max_rgb = max_rgb
        obj.max_a = max_a
        obj.linear = linear
        obj._format_params += ["max_rgb", "max_a", "linear"]
        return obj


class HSL(ColorPolarBase):
    """Represents a color in the HSL color space [#]_.

    .. [#] Wikipedia at https://en.wikipedia.org/wiki/HSL_and_HSV.

    Args:
        h: Hue component of the color.
        s: Saturation component of the color.
        l: Lightness component of the color.
        a: Opacity component of the color. Defaults to ``None``, which means it will be the same
            as the `max_sla` parameter.
        max_h: What is the maximum value for the `h` component. Some common values for this
            parameter would be 360 or 1.
        max_sla: What is the maximum value for the `s`, `l` and `a` components. Some common values
            for this parameter would be 1 or 100.
        include_a: Whether to include the opacity parameter `a` in the constructed tuple.
            Setting it to ``True`` may result in an object such as :code:`HSL(360, 1, 0,
            1)` instead of :code:`HSL(360, 1, 0)`, for exemple.
        round_to: Rounds the value of each color component to this many decimal places. Setting
            this parameter to 0 ensures that the components will be of type `int`. -1
            means that the components won't be rounded at all.
    """

    def __new__(cls,
                h: float,
                s: float,
                l: float,
                a: float = None,
                max_h=360,
                max_sla=1,
                include_a=False,
                round_to=-1):
        if a is None:
            a = max_sla
        if not all(0 <= spec <= max_sla for spec in (s, l, a)):
            raise ValueError("'s', 'l' and 'a' must be greater than 0 and smaller than 'max_sla'")

        rgba = colorsys.hls_to_rgb(h % max_h / max_h, l / max_sla, s / max_sla) + (a / max_sla,)

        obj = super().__new__(cls,
                              (h, s, l),
                              a,
                              np.array(rgba) * 255,
                              include_a=include_a,
                              round_to=round_to)
        obj.h, obj.s, obj.l = obj[:3]
        obj.max_h = max_h
        obj.max_sla = max_sla
        obj._format_params += ["max_h", "max_sla"]

        return obj

    @classmethod
    def _from_rgba(cls, rgba, max_h=360, max_sla=1, include_a=False, round_to=-1):
        hls = colorsys.rgb_to_hls(*np.array(rgba[:-1]) / 255)
        hsl = (hls[0] * max_h, hls[2] * max_sla, hls[1] * max_sla)

        obj = super().__new__(cls,
                              hsl,
                              rgba[-1] / 255 * max_sla,
                              rgba,
                              include_a=include_a,
                              round_to=round_to)
        obj.h, obj.s, obj.l = obj[:3]
        obj.max_h = max_h
        obj.max_sla = max_sla
        obj._format_params += ["max_h", "max_sla"]

        return obj


class HSV(ColorPolarBase):
    """Represents a color in the HSV color space [#]_.

    References:
        .. [#] Wikipedia at https://en.wikipedia.org/wiki/HSL_and_HSV.

    Args:
        h: Hue component of the color.
        s: Saturation component of the color.
        v: Value component of the color.
        a: Opacity component of the color. Defaults to ``None``, which means it will be the same
            as the `max_sva` parameter.
        max_h: What is the maximum value for the `h` component. Some common values for this
            parameter would be 360 or 1.
        max_sva: What is the maximum value for the `s`, `v` and `a` components. Some common values
            for this parameter would be 1 or 100.
        include_a: Whether to include the opacity parameter `a` in the constructed tuple.
            Setting it to ``True`` may result in an object such as :code:`HSV(360, 1, 0,
            1)` instead of :code:`HSV(360, 1, 0)`, for exemple.
        round_to: Rounds the value of each color component to this many decimal places. Setting
            this parameter to 0 ensures that the components will be of type `int`. -1
            means that the components won't be rounded at all.
    """

    def __new__(cls,
                h: float,
                s: float,
                v: float,
                a: float = None,
                max_h=360,
                max_sva=1,
                include_a=False,
                round_to=-1):
        if a is None:
            a = max_sva
        if not all(0 <= spec <= max_sva for spec in (s, v, a)):
            raise ValueError("'s', 'v' and 'a' must be greater than 0 and smaller than 'max_sva'")

        rgba = colorsys.hsv_to_rgb(h % max_h / max_h, s / max_sva, v / max_sva) + (a / max_sva,)

        obj = super().__new__(cls,
                              (h, s, v),
                              a,
                              np.array(rgba) * 255,
                              include_a=include_a,
                              round_to=round_to)
        obj.h, obj.s, obj.v = obj[:3]
        obj.max_h = max_h
        obj.max_sva = max_sva
        obj._format_params += ["max_h", "max_sva"]

        return obj

    @classmethod
    def _from_rgba(cls, rgba, max_h=360, max_sva=1, include_a=False, round_to=-1):
        hsv = colorsys.rgb_to_hsv(*np.array(rgba[:-1]) / 255)
        hsv = (hsv[0] * max_h, hsv[1] * max_sva, hsv[2] * max_sva)

        obj = super().__new__(cls,
                              hsv,
                              rgba[-1] / 255 * max_sva,
                              rgba,
                              include_a=include_a,
                              round_to=round_to)
        obj.h, obj.s, obj.v = obj[:3]
        obj.max_h = max_h
        obj.max_sva = max_sva
        obj._format_params += ["max_h", "max_sva"]

        return obj


class CMYK(ColorTupleBase):
    """Represents a color in the CMYK color space [#]_.

    References:
        .. [#] Wikipedia at https://en.wikipedia.org/wiki/CMYK_color_model.

    Args:
        c: Cyan component of the color.
        m: Magenta component of the color.
        y: Yellow component of the color.
        k: Key (black) component of the color.
        a: Opacity component of the color. Defaults to ``None``, which means it will be the same
            as the `max` parameter.
        max_cmyka: What is the maximum value for the `c`, `m`, `y`, `k` and `a` components. Some
            common values for this parameter would be 1 or 100.
        include_a: Whether to include the opacity parameter `a` in the constructed tuple.
            Setting it to ``True`` may result in an object such as :code:`CMYK(1, 1, 0,
            1)` instead of :code:`CMYK(1, 1, 0)`, for exemple.
        round_to: Rounds the value of each color component to this many decimal places. Setting
            this parameter to 0 ensures that the components will be of type `int`. The default,
            -1, means that the components won't be rounded at all.
    """

    def __new__(cls, c: float, m: float, y: float, k: float, a: float = None, max_cmyka=1,
                include_a=False,
                round_to=-1):
        if a is None:
            a = max_cmyka
        if not all(0 <= spec <= max_cmyka for spec in (c, m, y, k, a)):
            raise ValueError("'c', 'm', 'y', 'k', and 'a' must be greater than 0 and smaller than "
                             "'max_cmyka'")

        rgba = convert_color(
            CMYKColor(*(np.array((c, m, y, k)) / max_cmyka)),
            sRGBColor
        ).get_value_tuple() + (a / max_cmyka,)

        obj = super().__new__(cls,
                              (c, m, y, k),
                              a,
                              np.array(rgba) * 255,
                              include_a=include_a,
                              round_to=round_to)
        obj.c, obj.m, obj.y, obj.k = obj[:4]
        obj.max_cmyka = max_cmyka
        obj._format_params.append("max_cmyka")

        return obj

    @classmethod
    def _from_rgba(cls, rgba, max_cmyka=1, include_a=False, round_to=-1):
        cmyk = convert_color(
            sRGBColor(*rgba[:3], is_upscaled=True),
            CMYKColor
        ).get_value_tuple()
        cmyk = np.array(cmyk) * max_cmyka

        obj = super().__new__(cls,
                              cmyk,
                              rgba[-1] / 255 * max_cmyka,
                              rgba,
                              include_a=include_a,
                              round_to=round_to)
        obj.c, obj.m, obj.y, obj.k = obj[:4]
        obj.max_cmyka = max_cmyka
        obj._format_params.append("max_cmyka")

        return obj


class CMY(ColorTupleBase):
    """Represents a color in the CMY color space [#]_.

    References:
        .. [#] Wikipedia at https://en.wikipedia.org/wiki/CMY_color_model.


    Args:
        c: Cyan component of the color.
        m: Magenta component of the color.
        y: Yellow component of the color.
        a: Opacity component of the color. Defaults to ``None``, which means it will be the same
            as the `max` parameter.
        max_cmya: What is the maximum value for the `c`, `m`, `y` and `a` components. Some common
            values for this parameter would be 1 or 100.
        include_a: Whether to include the opacity parameter `a` in the constructed tuple.
            Setting it to ``True`` may result in an object such as :code:`CMY(1, 1, 0,
            1)` instead of :code:`CMY(1, 1, 0)`, for exemple.
        round_to: Rounds the value of each color component to this many decimal places. Setting
            this parameter to 0 ensures that the components will be of type `int`. The default,
            -1, means that the components won't be rounded at all.
    """

    def __new__(cls,
                c: float,
                m: float,
                y: float,
                a: float = None,
                max_cmya=1,
                include_a=False,
                round_to=-1):
        if a is None:
            a = max_cmya
        if not all(0 <= spec <= max_cmya for spec in (c, m, y, a)):
            raise ValueError("'c', 'm', 'y', and 'a' must be greater than 0 and smaller than "
                             "'max_cmya'")

        rgba = convert_color(
            CMYColor(*(np.array((c, m, y)) / max_cmya)),
            sRGBColor
        ).get_value_tuple() + (a / max_cmya,)

        obj = super().__new__(
            cls,
            (c, m, y),
            a,
            np.array(rgba) * 255,
            include_a=include_a,
            round_to=round_to)
        obj.c, obj.m, obj.y = obj[:3]
        obj.max_cmya = max_cmya
        obj._format_params.append("max_cmya")

        return obj

    @classmethod
    def _from_rgba(cls, rgba, max_cmya=1, include_a=False, round_to=-1):
        cmy = convert_color(
            sRGBColor(*rgba[:3], is_upscaled=True),
            CMYColor
        ).get_value_tuple()
        cmy = np.array(cmy) * max_cmya

        obj = super().__new__(cls,
                              cmy,
                              rgba[-1] / 255 * max_cmya,
                              rgba,
                              include_a=include_a,
                              round_to=round_to)
        obj.c, obj.m, obj.y = obj[:3]
        obj.max_cmya = max_cmya
        obj._format_params.append("max_cmya")

        return obj


# TODO doc
class CIELuv(ColorTupleBase):
    def __new__(cls, l, u, v, a=None, max_a=1, include_a=False, round_to=-1):
        if a is None:
            a = max_a
        elif not 0 <= a <= max_a:
            raise ValueError("'a' must be greater than 0 and smaller than 'max_a'")
        if not 0 <= l <= 100:
            raise ValueError("'l' must be greater than 0 and smaller than 100")

        rgb = convert_color(LuvColor(l, u, v, illuminant="d65"), sRGBColor)
        rgba = (rgb.clamped_rgb_r * 255,
                rgb.clamped_rgb_g * 255,
                rgb.clamped_rgb_b * 255,
                a / max_a * 255)
        obj = super().__new__(cls, (l, u, v), a, rgba, include_a=include_a, round_to=round_to)
        obj.l, obj.u, obj.v = obj[:3]
        obj.max_a = max_a
        obj._format_params.append("max_a")

        return obj

    @classmethod
    def _from_rgba(cls, rgba, max_a=1, include_a=False, round_to=-1):
        luv = convert_color(
            sRGBColor(*rgba[:3], is_upscaled=True),
            LuvColor,
            target_illuminant="d65"
        ).get_value_tuple()
        obj = super().__new__(cls,
                              luv,
                              rgba[-1] / 255 * max_a,
                              rgba,
                              include_a=include_a,
                              round_to=round_to)
        obj.l, obj.u, obj.v = obj[:3]
        obj.max_a = max_a
        obj._format_params.append("max_a")

        return obj


# TODO doc
class CIELab(ColorTupleBase):
    def __new__(cls, l, a_, b, a=None, max_a=1, include_a=False, round_to=-1):
        if a is None:
            a = max_a
        elif not 0 <= a <= max_a:
            raise ValueError("'a' must be greater than 0 and smaller than 'max_a'")
        if not 0 <= l <= 100:
            raise ValueError("'l' must be greater than 0 and smaller than 100")

        rgb = convert_color(LabColor(l, a_, b, illuminant="d65"), sRGBColor)
        rgba = (rgb.clamped_rgb_r * 255,
                rgb.clamped_rgb_g * 255,
                rgb.clamped_rgb_b * 255,
                a / max_a * 255)
        obj = super().__new__(cls, (l, a_, b), a, rgba, include_a=include_a, round_to=round_to)
        obj.l, obj.a_, obj.b = l, a_, b
        obj.max_a = max_a
        obj._format_params.append("max_a")

        return obj

    @classmethod
    def _from_rgba(cls, rgba, max_a=1, include_a=False, round_to=-1):
        lab = convert_color(
            sRGBColor(*rgba[:3], is_upscaled=True),
            LabColor,
            target_illuminant="d65"
        ).get_value_tuple()

        obj = super().__new__(cls,
                              lab,
                              rgba[-1] / 255 * max_a,
                              rgba,
                              include_a=include_a,
                              round_to=round_to)
        obj.l, obj.a_, obj.b = obj[:3]
        obj.max_a = max_a
        obj._format_params.append("max_a")

        return obj


# TODO doc
class HCLuv(ColorPolarBase):
    def __new__(cls, h, c, l, a=None, max_h=360, max_a=1, include_a=False, round_to=-1):
        if a is None:
            a = max_a
        elif not 0 <= a <= max_a:
            raise ValueError("'a' must be greater than 0 and smaller than 'max_a'")
        if not 0 <= l <= 100:
            raise ValueError("'l' must be greater than 0 and smaller than 100")

        rgb = convert_color(
            LCHuvColor(l, c, h / max_h * 360 % 360, illuminant="d65"),
            sRGBColor
        )
        rgba = (rgb.clamped_rgb_r * 255,
                rgb.clamped_rgb_g * 255,
                rgb.clamped_rgb_b * 255,
                a / max_a * 255)
        obj = super().__new__(cls,
                              (h, c, l),
                              a,
                              rgba,
                              include_a=include_a,
                              round_to=round_to)
        obj.h, obj.c, obj.l = obj[:3]
        obj.max_h = max_h
        obj.max_a = max_a
        obj._format_params += ["max_h", "max_a"]

        return obj

    @classmethod
    def _from_rgba(cls, rgba, max_h=360, max_a=1, include_a=False, round_to=-1):
        hcl = convert_color(
            sRGBColor(*rgba[:3], is_upscaled=True),
            LCHuvColor,
            target_illuminant="d65"
        ).get_value_tuple()[::-1]
        hcl = np.array(hcl)
        hcl[0] *= max_h / 360
        obj = super().__new__(
            cls,
            hcl,
            rgba[-1] / 255 * max_a,
            rgba,
            include_a=include_a,
            round_to=round_to
        )
        obj.h, obj.c, obj.l = obj[:3]
        obj.max_h = max_h
        obj.max_a = max_a
        obj._format_params += ["max_h", "max_a"]

        return obj


# TODO doc
class HCLab(ColorPolarBase):
    def __new__(cls, h, c, l, a=None, max_h=360, max_a=1, include_a=False, round_to=-1):
        if a is None:
            a = max_a
        elif not 0 <= a <= max_a:
            raise ValueError("'a' must be greater than 0 and smaller than 'max_a'")
        if not 0 <= l <= 100:
            raise ValueError("'l' must be greater than 0 and smaller than 100")

        rgb = convert_color(
            LCHabColor(l, c, h / max_h * 360 % 360, illuminant="d65"),
            sRGBColor
        )
        rgba = (rgb.clamped_rgb_r * 255,
                rgb.clamped_rgb_g * 255,
                rgb.clamped_rgb_b * 255,
                a / max_a * 255)
        obj = super().__new__(cls,
                              (h, c, l),
                              a,
                              rgba,
                              include_a=include_a,
                              round_to=round_to)
        obj.h, obj.c, obj.l = obj[:3]
        obj.max_h = max_h
        obj.max_a = max_a
        obj._format_params += ["max_h", "max_a"]

        return obj

    @classmethod
    def _from_rgba(cls, rgba, max_h=360, max_a=1, include_a=False, round_to=-1):
        hcl = convert_color(
            sRGBColor(*rgba[:3], is_upscaled=True),
            LCHabColor,
            target_illuminant="d65"
        ).get_value_tuple()[::-1]
        hcl = np.array(hcl)
        hcl[0] *= max_h / 360
        obj = super().__new__(
            cls,
            hcl,
            rgba[-1] / 255 * max_a,
            rgba,
            include_a=include_a,
            round_to=round_to
        )
        obj.h, obj.c, obj.l = obj[:3]
        obj.max_h = max_h
        obj.max_a = max_a
        obj._format_params += ["max_h", "max_a"]

        return obj


default_tail = object()


def _warn_tail():
    warnings.warn("the default value for the 'tail_a' argument will be changed to 'True' "
                  "in the next release. Pass the argument explicitly with 'Hex(string, tail_a=False)' instead.",
                  FutureWarning,
                  stacklevel=3)


class Hex(ColorBase, str):
    """Represents a color in the RGB color space [#]_ as a hexadecimal string.

    Is mostly used for representing colors in web applications [#]_.

    References:
        .. [#] Wikipedia at https://en.wikipedia.org/wiki/SRGB.
        .. [#] Wikipedia at https://en.wikipedia.org/wiki/Web_colors

    Args:
        hex_str: Hexadecimal string from which the :class:`Hex` instance will be built.
            The format can be a string of 6 digits, 8 digits (with an alpha specifier in the beginning or the end),
            or 3 digits, optionally including a '#' character.
        uppercase: Whether the color will be represented in uppercase or lowercase.
        include_a: Whether to include the opacity parameter `a` in the constructed string.
            Setting it to ``True`` may result in an object such as :code:`Hex('#ffffff00')`
            instead of :code:`Hex('#ffff00')`, for exemple.
        tail_a: Whether the alpha component is present in the tail or head of the hex string. Used
            only if `hex_str` includes an alpha component or `include_a` is ``True``.

    Examples:
        >>> Hex("#ff0000")
        Hex('#ff0000')

        >>> Hex("#FF0000", include_hash=False)
        Hex('ff0000')

        >>> Hex("ff0000", include_a=True, tail_a=True)
        Hex('#ff0000ff')
    """

    def __new__(cls,
                hex_str: str,
                uppercase=False,
                include_hash=True,
                include_a=False,
                tail_a=default_tail):
        if tail_a is default_tail:
            tail_a = False
            if include_a or len(hex_str) >= 8:
                _warn_tail()

        hex_str = hex_str.lstrip("#")
        if len(hex_str) not in (3, 6, 8):
            raise ValueError("'hex_str' length must be 3, 6 or 8 (excluding the optional '#')")
        if len(hex_str) == 3:
            hex_str = "".join(i + j for i, j in zip(hex_str, hex_str))
        if len(hex_str) < 8:
            i = 0
            a = 255
        elif tail_a:
            i = 0
            a = int(hex_str[-2:], 16)
        else:
            i = 2
            a = int(hex_str[:2], 16)
        rgb = (int(hex_str[i:i + 2], 16),
               int(hex_str[i + 2:i + 4], 16),
               int(hex_str[i + 4:i + 6], 16))
        # Rather than modifying the stored string to suit the format specifications we just
        # delegate that to the _from_rgba method
        return cls._from_rgba(rgb + (a,),
                              uppercase=uppercase,
                              include_hash=include_hash,
                              include_a=include_a,
                              tail_a=tail_a)

    @classmethod
    def _from_rgba(cls, rgba, uppercase=False, include_hash=True, include_a=False, tail_a=default_tail):
        if tail_a is default_tail:
            tail_a = False
            if include_a:
                _warn_tail()

        rgba = tuple(rgba)
        hex_str = '%02x%02x%02x' % rgba[:3]

        if include_a:
            a = '%02x' % rgba[-1]
            if not tail_a:
                hex_str = a + hex_str
            else:
                hex_str += a
        if uppercase:
            hex_str = hex_str.upper()
        if include_hash:
            hex_str = '#' + hex_str

        obj = str.__new__(cls, hex_str)
        obj.uppercase = uppercase
        obj.include_hash = include_hash
        obj.include_a = include_a
        obj.tail_a = tail_a
        obj._rgba = np.rint(rgba).astype(int)
        obj._format_params = ["uppercase", "include_hash", "include_a", "tail_a"]
        return obj

    # It would be dangerous to change str conversion as the target framework could call it
    # expecting #ff0000 and get Hex('#ff0000')
    def __str__(self):
        return str.__str__(self)

    def __repr__(self):
        if colorir.config.REPR_STYLE == "traditional":
            return f"{self.__class__.__name__}({str.__repr__(self)})"
        if colorir.config.REPR_STYLE == "inherit":
            return str(self)
        return colorir.utils.swatch(self, file=None)

    def __hash__(self):
        return ColorBase.__hash__(self)

    def __eq__(self, other):
        colorbase_eq = ColorBase.__eq__(self, other)
        # If other is ColorBase than we trust the result of eq
        if isinstance(other, ColorBase):
            return colorbase_eq
        # Otherwise we also try str.__eq__
        return any([colorbase_eq is True, str.__eq__(self, other) is True])


# Aliases
HexRGB = Hex
sRGB = RGB

ColorLike = Union[ColorBase, str, tuple]
"""Type constant that describes common representations of colors in python."""



# -*- coding: utf-8 -*-
import itertools
import numpy as np
import unittest
from colorir.colormath import color_conversions
from colorir.colormath.color_conversions import (
    GraphConversionManager,
    XYZ_to_RGB,
    HSV_to_RGB,
    RGB_to_XYZ,
)
from colorir.colormath.color_exceptions import UndefinedConversionError
from colorir.colormath.color_objects import (
    XYZColor,
    BaseRGBColor,
    HSVColor,
    HSLColor,
    AdobeRGBColor,
    BT2020Color,
    sRGBColor,
)


class GraphConversionManagerTestCase(unittest.TestCase):
    def setUp(self):
        self.manager = GraphConversionManager()
        self.manager.add_type_conversion(XYZColor, BaseRGBColor, XYZ_to_RGB)
        self.manager.add_type_conversion(BaseRGBColor, HSVColor, HSV_to_RGB)

    def test_basic_path_generation(self):
        path = self.manager.get_conversion_path(XYZColor, HSVColor)
        self.assertEqual(path, [XYZ_to_RGB, HSV_to_RGB])

    def test_self_conversion(self):
        path = self.manager.get_conversion_path(XYZColor, XYZColor)
        self.assertEqual(path, [])

    def test_invalid_path_response(self):
        self.assertRaises(
            UndefinedConversionError,
            self.manager.get_conversion_path,
            XYZColor,
            HSLColor,
        )


class ColorConversionTestCase(unittest.TestCase):
    def test_conversion_validity(self):
        """
        Make sure that all existing paths between colors are valid.
        """

        conversion_manager = color_conversions._conversion_manager
        color_spaces = conversion_manager.registered_color_spaces

        for start_space, target_space in itertools.product(color_spaces, repeat=2):
            try:
                path = conversion_manager.get_conversion_path(start_space, target_space)
            except UndefinedConversionError:
                # If there is no path we don't really care (e.g., conversion to
                # SpectralColor).
                continue

            if start_space == target_space:
                # If start and end color space are equal, the conversion path should
                # be empty.
                self.assertEqual(path, [])
                continue
            else:
                # Otherwise check that all the conversion functions math up
                for a, b in zip(path[:-1], path[1:]):
                    self.assertEqual(a.target_type, b.start_type)

    def test_transfer_functions(self):
        """
        Tests the transfer functions of the various RGB colorspaces.
        """

        for colorspace in (AdobeRGBColor, BT2020Color, sRGBColor):
            for a in (0.0, 0.01, 0.18, 1.0):
                RGB = [a] * 3
            np.testing.assert_allclose(
                XYZ_to_RGB(RGB_to_XYZ(colorspace(*RGB)), colorspace).get_value_tuple(),
                RGB,
                rtol=1e-5,
                atol=1e-5,
            )



import doctest
import unittest

from colorir import *

config.REPR_STYLE = "traditional"


def load_tests(loader, tests, ignore):
    tests.addTests(doctest.DocTestSuite(gradient))
    return tests


if __name__ == "__main__":
    unittest.main()


# -*- coding: utf-8 -*-
"""
This module contains the formulas for comparing Lab values with matrices
and vectors. The benefit of using NumPy's matrix capabilities is speed. These
calls can be used to efficiently compare large volumes of Lab colors.
"""

import numpy


def delta_e_cie1976(lab_color_vector, lab_color_matrix):
    """
    Calculates the Delta E (CIE1976) between `lab_color_vector` and all
    colors in `lab_color_matrix`.
    """
    return numpy.sqrt(
        numpy.sum(numpy.power(lab_color_vector - lab_color_matrix, 2), axis=1)
    )


# noinspection PyPep8Naming
def delta_e_cie1994(
    lab_color_vector, lab_color_matrix, K_L=1, K_C=1, K_H=1, K_1=0.045, K_2=0.015
):
    """
    Calculates the Delta E (CIE1994) of two colors.

    K_l:
      0.045 graphic arts
      0.048 textiles
    K_2:
      0.015 graphic arts
      0.014 textiles
    K_L:
      1 default
      2 textiles
    """
    C_1 = numpy.sqrt(numpy.sum(numpy.power(lab_color_vector[1:], 2)))
    C_2 = numpy.sqrt(numpy.sum(numpy.power(lab_color_matrix[:, 1:], 2), axis=1))

    delta_lab = lab_color_vector - lab_color_matrix

    delta_L = delta_lab[:, 0].copy()
    delta_C = C_1 - C_2
    delta_lab[:, 0] = delta_C

    delta_H_sq = numpy.sum(numpy.power(delta_lab, 2) * numpy.array([-1, 1, 1]), axis=1)
    # noinspection PyArgumentList
    delta_H = numpy.sqrt(delta_H_sq.clip(min=0))

    S_L = 1
    S_C = 1 + K_1 * C_1
    S_H = 1 + K_2 * C_1

    LCH = numpy.vstack([delta_L, delta_C, delta_H])
    params = numpy.array([[K_L * S_L], [K_C * S_C], [K_H * S_H]])

    return numpy.sqrt(numpy.sum(numpy.power(LCH / params, 2), axis=0))


# noinspection PyPep8Naming
def delta_e_cmc(lab_color_vector, lab_color_matrix, pl=2, pc=1):
    """
    Calculates the Delta E (CIE1994) of two colors.

    CMC values
      Acceptability: pl=2, pc=1
      Perceptability: pl=1, pc=1
    """
    L, a, b = lab_color_vector

    C_1 = numpy.sqrt(numpy.sum(numpy.power(lab_color_vector[1:], 2)))
    C_2 = numpy.sqrt(numpy.sum(numpy.power(lab_color_matrix[:, 1:], 2), axis=1))

    delta_lab = lab_color_vector - lab_color_matrix

    delta_L = delta_lab[:, 0].copy()
    delta_C = C_1 - C_2
    delta_lab[:, 0] = delta_C

    H_1 = numpy.degrees(numpy.arctan2(b, a))

    if H_1 < 0:
        H_1 += 360

    F = numpy.sqrt(numpy.power(C_1, 4) / (numpy.power(C_1, 4) + 1900.0))

    # noinspection PyChainedComparisons
    if 164 <= H_1 and H_1 <= 345:
        T = 0.56 + abs(0.2 * numpy.cos(numpy.radians(H_1 + 168)))
    else:
        T = 0.36 + abs(0.4 * numpy.cos(numpy.radians(H_1 + 35)))

    if L < 16:
        S_L = 0.511
    else:
        S_L = (0.040975 * L) / (1 + 0.01765 * L)

    S_C = ((0.0638 * C_1) / (1 + 0.0131 * C_1)) + 0.638
    S_H = S_C * (F * T + 1 - F)

    delta_C = C_1 - C_2

    delta_H_sq = numpy.sum(numpy.power(delta_lab, 2) * numpy.array([-1, 1, 1]), axis=1)
    # noinspection PyArgumentList
    delta_H = numpy.sqrt(delta_H_sq.clip(min=0))

    LCH = numpy.vstack([delta_L, delta_C, delta_H])
    params = numpy.array([[pl * S_L], [pc * S_C], [S_H]])

    return numpy.sqrt(numpy.sum(numpy.power(LCH / params, 2), axis=0))


# noinspection PyPep8Naming
def delta_e_cie2000(lab_color_vector, lab_color_matrix, Kl=1, Kc=1, Kh=1):
    """
    Calculates the Delta E (CIE2000) of two colors.
    """
    L, a, b = lab_color_vector

    avg_Lp = (L + lab_color_matrix[:, 0]) / 2.0

    C1 = numpy.sqrt(numpy.sum(numpy.power(lab_color_vector[1:], 2)))
    C2 = numpy.sqrt(numpy.sum(numpy.power(lab_color_matrix[:, 1:], 2), axis=1))

    avg_C1_C2 = (C1 + C2) / 2.0

    G = 0.5 * (
        1
        - numpy.sqrt(
            numpy.power(avg_C1_C2, 7.0)
            / (numpy.power(avg_C1_C2, 7.0) + numpy.power(25.0, 7.0))
        )
    )

    a1p = (1.0 + G) * a
    a2p = (1.0 + G) * lab_color_matrix[:, 1]

    C1p = numpy.sqrt(numpy.power(a1p, 2) + numpy.power(b, 2))
    C2p = numpy.sqrt(numpy.power(a2p, 2) + numpy.power(lab_color_matrix[:, 2], 2))

    avg_C1p_C2p = (C1p + C2p) / 2.0

    h1p = numpy.degrees(numpy.arctan2(b, a1p))
    h1p += (h1p < 0) * 360

    h2p = numpy.degrees(numpy.arctan2(lab_color_matrix[:, 2], a2p))
    h2p += (h2p < 0) * 360

    avg_Hp = (((numpy.fabs(h1p - h2p) > 180) * 360) + h1p + h2p) / 2.0

    T = (
        1
        - 0.17 * numpy.cos(numpy.radians(avg_Hp - 30))
        + 0.24 * numpy.cos(numpy.radians(2 * avg_Hp))
        + 0.32 * numpy.cos(numpy.radians(3 * avg_Hp + 6))
        - 0.2 * numpy.cos(numpy.radians(4 * avg_Hp - 63))
    )

    diff_h2p_h1p = h2p - h1p
    delta_hp = diff_h2p_h1p + (numpy.fabs(diff_h2p_h1p) > 180) * 360
    delta_hp -= (h2p > h1p) * 720

    delta_Lp = lab_color_matrix[:, 0] - L
    delta_Cp = C2p - C1p
    delta_Hp = 2 * numpy.sqrt(C2p * C1p) * numpy.sin(numpy.radians(delta_hp) / 2.0)

    S_L = 1 + (
        (0.015 * numpy.power(avg_Lp - 50, 2))
        / numpy.sqrt(20 + numpy.power(avg_Lp - 50, 2.0))
    )
    S_C = 1 + 0.045 * avg_C1p_C2p
    S_H = 1 + 0.015 * avg_C1p_C2p * T

    delta_ro = 30 * numpy.exp(-(numpy.power(((avg_Hp - 275) / 25), 2.0)))
    R_C = numpy.sqrt(
        (numpy.power(avg_C1p_C2p, 7.0))
        / (numpy.power(avg_C1p_C2p, 7.0) + numpy.power(25.0, 7.0))
    )
    R_T = -2 * R_C * numpy.sin(2 * numpy.radians(delta_ro))

    return numpy.sqrt(
        numpy.power(delta_Lp / (S_L * Kl), 2)
        + numpy.power(delta_Cp / (S_C * Kc), 2)
        + numpy.power(delta_Hp / (S_H * Kh), 2)
        + R_T * (delta_Cp / (S_C * Kc)) * (delta_Hp / (S_H * Kh))
    )



# -*- coding: utf-8 -*-
"""
Copyright (c) 2014, Michael Mauderer, University of St Andrews
All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

 * Redistributions of source code must retain the above copyright notice, this
   list of conditions and the following disclaimer.
 * Redistributions in binary form must reproduce the above copyright notice,
   this list of conditions and the following disclaimer in the documentation
   and/or other materials provided with the distribution.
 * Neither the name of the University of St Andrews nor the names of its
   contributors may be used to endorse or promote products derived from this
   software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""

from abc import abstractmethod
from collections import defaultdict
import csv
import os

import unittest
import numpy

from numpy.testing import assert_allclose, assert_almost_equal
from colorir.colormath.color_appearance_models import (
    CIECAM02,
    RLAB,
    LLAB,
    ATD95,
    Nayatani95,
    Hunt,
)


class ColorAppearanceTest(object):
    fixture_path = None
    output_parameter_dict = {}

    @staticmethod
    def load_fixture(file_name):
        path = os.path.dirname(__file__)
        with open(os.path.join(path, "colormath_fixtures", file_name)) as in_file:
            result = []
            for case_data in csv.DictReader(in_file):
                for key in case_data:
                    try:
                        case_data[key] = float(case_data[key])
                    except ValueError:
                        pass
                result.append(case_data)
            return result

    def check_model_consistency(self, data, output_parameter_dict):
        for data_attr, model_attr in sorted(output_parameter_dict.items()):
            yield self.check_model_attribute, data.get("Case"), data, model_attr, data[
                data_attr
            ]

    @abstractmethod
    def create_model_from_data(self, data):
        pass

    def check_model_attribute(self, case, data, model_attr, target):
        model = self.create_model_from_data(data)
        model_parameter = getattr(model, model_attr)
        error_message = (
            "Parameter {} in test case {} does not match target value."
            "\nExpected: {} \nReceived {}"
        ).format(model_attr, case, target, model_parameter)

        assert_allclose(
            model_parameter,
            target,
            err_msg=error_message,
            rtol=0.01,
            atol=0.01,
            verbose=False,
        )
        assert_almost_equal(model_parameter, target, decimal=1, err_msg=error_message)

    limited_fixtures = None

    def _get_fixtures(self):
        # Sometimes it might be desirable to exclude s specific fixture for testing
        fixtures = self.load_fixture(self.fixture_path)
        if self.limited_fixtures is not None:
            fixtures = [fixtures[index] for index in self.limited_fixtures]
        return fixtures

    def test_forward_examples(self):
        # Go through all available fixtures
        for data in self._get_fixtures():
            # Create a single test for each output parameter
            for test in self.check_model_consistency(data, self.output_parameter_dict):
                yield test

    def test_parallel_forward_example(self):
        # Collect all fixture data in a single dict of lists
        data = defaultdict(list)
        for fixture in self._get_fixtures():
            for key, value in fixture.items():
                data[key].append(value)
        # Turn lists into numpy.arrays
        for key in data:
            data[key] = numpy.array(data[key])
        # Create tests
        for test in self.check_model_consistency(data, self.output_parameter_dict):
            yield test


class TestNayataniColorAppearanceModel(ColorAppearanceTest):
    fixture_path = "nayatani.csv"

    output_parameter_dict = {
        "L_star_P": "_lightness_achromatic",
        "L_star_N": "_lightness_achromatic_normalized",
        "theta": "hue_angle",
        "C": "chroma",
        "S": "saturation",
        "B_r": "brightness",
        "M": "colorfulness",
    }

    def create_model_from_data(self, data):
        model = Nayatani95(
            data["X"],
            data["Y"],
            data["Z"],
            data["X_n"],
            data["Y_n"],
            data["Z_n"],
            data["Y_o"],
            data["E_o"],
            data["E_or"],
        )
        return model

    @staticmethod
    def test_beta_1():
        assert_almost_equal(Nayatani95._beta_1(0), 1, decimal=6)
        assert_almost_equal(Nayatani95._beta_1(1), 1.717900656, decimal=6)
        assert_almost_equal(Nayatani95._beta_1(2), 1.934597896, decimal=6)

    @staticmethod
    def test_beta_2():
        assert_almost_equal(Nayatani95._beta_2(0), 0.7844, decimal=6)
        assert_almost_equal(Nayatani95._beta_2(1), 1.375241343, decimal=6)
        assert_almost_equal(Nayatani95._beta_2(2), 1.59085866, decimal=6)


class TestHuntColorAppearanceModel(ColorAppearanceTest):
    fixture_path = "hunt.csv"

    output_parameter_dict = {
        "h_S": "hue_angle",
        "s": "saturation",
        "Q": "brightness",
        "J": "lightness",
        "C_94": "chroma",
        "M94": "colorfulness",
    }

    def create_model_from_data(self, data):
        return Hunt(
            data["X"],
            data["Y"],
            data["Z"],
            data["X_W"],
            0.2 * data["Y_W"],
            data["Z_W"],
            data["X_W"],
            data["Y_W"],
            data["Z_W"],
            l_a=data["L_A"],
            n_c=data["N_c"],
            n_b=data["N_b"],
            cct_w=data["T"],
        )


class TestRLABColorAppearanceModel(ColorAppearanceTest):
    fixture_path = "rlab.csv"
    output_parameter_dict = {
        "L": "lightness",
        "C": "chroma",
        "s": "saturation",
        "a": "a",
        "b": "b",
        "h": "hue_angle",
    }

    def create_model_from_data(self, data):
        model = RLAB(
            data["X"],
            data["Y"],
            data["Z"],
            data["X_n"],
            data["Y_n"],
            data["Z_n"],
            data["Y_n2"],
            data["sigma"],
            data["D"],
        )
        return model


class TestATDColorAppearanceModel(ColorAppearanceTest):
    fixture_path = "atd.csv"

    output_parameter_dict = {
        "A_1": "_a_1",
        "T_1": "_t_1",
        "D_1": "_d_1",
        "A_2": "_a_2",
        "T_2": "_t_2",
        "D_2": "_d_2",
        "Br": "brightness",
        "C": "saturation",
        "H": "hue",
    }

    def create_model_from_data(self, data):
        model = ATD95(
            data["X"],
            data["Y"],
            data["Z"],
            data["X_0"],
            data["Y_0"],
            data["Z_0"],
            data["Y_02"],
            data["K_1"],
            data["K_2"],
            data["sigma"],
        )
        return model

    @staticmethod
    def test_xyz_to_lms():
        l, m, s = ATD95._xyz_to_lms(numpy.array([1, 1, 1]))
        assert_almost_equal(l, 0.7946522478109985)
        assert_almost_equal(m, 0.9303058494144267)
        assert_almost_equal(s, 0.7252006614718631)

    @staticmethod
    def test_final_response_calculation():
        assert_almost_equal(ATD95._calculate_final_response(0), 0)
        assert_almost_equal(ATD95._calculate_final_response(100), 1.0 / 3.0)
        assert_almost_equal(ATD95._calculate_final_response(200), 0.5)
        assert_almost_equal(ATD95._calculate_final_response(10000), 0.980392157)


class TestLLABColorAppearanceModel(ColorAppearanceTest):
    fixture_path = "llab.csv"

    output_parameter_dict = {
        "L_L": "lightness",
        "Ch_L": "chroma",
        "s_L": "saturation",
        "h_L": "hue_angle",
        "A_L": "a_l",
        "B_L": "b_l",
    }

    def create_model_from_data(self, data):
        return LLAB(
            data["X"],
            data["Y"],
            data["Z"],
            data["X_0"],
            data["Y_0"],
            data["Z_0"],
            data["Y_b"],
            data["F_S"],
            data["F_L"],
            data["F_C"],
            data["L"],
        )


class TestCIECAM02ColorAppearanceModel(ColorAppearanceTest):
    fixture_path = "ciecam02.csv"
    output_parameter_dict = {
        "J": "lightness",
        "Q": "brightness",
        "C": "chroma",
        "M": "colorfulness",
        "S": "saturation",
        "N_bb": "n_bb",
        "a_c": "a_c",
        "b_c": "b_c",
        "a_M": "a_m",
        "b_M": "b_m",
        "a_s": "a_s",
        "b_s": "b_s",
    }

    input_parameter_dict = {}

    def create_model_from_data(self, data):
        return CIECAM02(
            data["X"],
            data["Y"],
            data["Z"],
            data["X_W"],
            data["Y_W"],
            data["Z_W"],
            data["Y_b"],
            data["L_A"],
            data["c"],
            data["N_c"],
            data["F"],
        )

    @staticmethod
    def test_degree_of_adaptation():
        assert_almost_equal(
            CIECAM02._compute_degree_of_adaptation(1, 100), 0.940656, decimal=6
        )
        assert_almost_equal(
            CIECAM02._compute_degree_of_adaptation(1, 318.31), 0.9994, decimal=2
        )
        assert_almost_equal(
            CIECAM02._compute_degree_of_adaptation(1, 31.83), 0.875, decimal=2
        )

    @staticmethod
    def test_xyz_to_rgb():
        assert_almost_equal(
            CIECAM02._xyz_to_rgb(numpy.array([95.05, 100, 108.9])),
            numpy.array([94.9273, 103.527, 108.737]),
            decimal=2,
        )


@unittest.skip
class TestCIECAM02m1ColorAppearanceModel(ColorAppearanceTest):
    # TODO: Test CIECAM02-m1 model
    pass



import tkinter as tk

import numpy as np

from . import config
from .utils import color_str


class StaticPaletteWidget(tk.Canvas):
    def __init__(self, master, colors, width, height):
        super().__init__(master, width=width, height=height)

        rect_w = width / len(colors)
        for i, color in enumerate(colors):
            x = i * rect_w
            color = config.DEFAULT_COLOR_FORMAT.format(color).hex()
            self.create_rectangle(
                x, 0, x + rect_w, height,
                fill=color,
                outline=color
            )


class ColorFrame(tk.Frame):
    def __init__(self, master, root, color, color_name, width, height):
        super().__init__(master, width=width, height=height)

        self.root = root
        self.color = config.DEFAULT_COLOR_FORMAT.format(color)
        self.color_name = color_name
        self.btn = tk.Button(self,
                             command=self.on_click,
                             bg=self.color.hex(),
                             activebackground=self.color.hex(),
                             bd=0,
                             highlightthickness=0)
        self.btn.pack(fill=tk.BOTH, expand=1)
        self.pack_propagate(0)

    def on_click(self):
        c_str = color_str(str(self.color), self.color)
        if self.color_name:
            c_str = self.color_name + " " + c_str
        c_str = f"{color_str('   ', bg_color=self.color)} {color_str(c_str, self.color)}"
        print(c_str)
        # Transmit clipboard command to PaletteWidget
        self.send_clipboard(str(self.color))

    def send_clipboard(self, string):
        self.root.clipboard_clear()
        self.root.clipboard_append(string)
        # Prevents from erasing clipboard after thread is closed
        self.root.update()


class PaletteWidget(tk.Frame):
    def __init__(self, master, colors, width, height, color_names=None):
        super().__init__(master, width=width, height=height)

        root = self.get_root()
        if color_names is None:
            color_names = [""] * len(colors)
        if len(colors) > width:
            indexes = np.linspace(0, len(colors) - 1, width).round().astype(int)
            colors = [colors[i] for i in indexes]

        btn_w = int(width / len(colors))
        for color, color_name in zip(colors, color_names):
            btn = ColorFrame(self, root, color, color_name, btn_w, height)
            btn.pack(side="left")

    def get_root(self):
        wgt = self
        while wgt.master is not None:
            wgt = wgt.master
        return wgt




# -*- coding: utf-8 -*-
"""
Conversion between color spaces.

.. note:: This module makes extensive use of imports within functions.
    That stinks.
"""
from abc import ABCMeta, abstractmethod

import math
import logging

import numpy
import networkx

from . import color_constants
from . import spectral_constants
from .color_objects import (
    ColorBase,
    XYZColor,
    sRGBColor,
    LCHabColor,
    LCHuvColor,
    LabColor,
    xyYColor,
    LuvColor,
    HSVColor,
    HSLColor,
    CMYColor,
    CMYKColor,
    BaseRGBColor,
    IPTColor,
    SpectralColor,
    BT2020Color,
)
from .chromatic_adaptation import apply_chromatic_adaptation
from .color_exceptions import InvalidIlluminantError, UndefinedConversionError


logger = logging.getLogger(__name__)


# noinspection PyPep8Naming
def apply_RGB_matrix(var1, var2, var3, rgb_type, convtype="xyz_to_rgb"):
    """
    Applies an RGB working matrix to convert from XYZ to RGB.
    The arguments are tersely named var1, var2, and var3 to allow for the
    passing of XYZ _or_ RGB values. var1 is X for XYZ, and R for RGB. var2 and
    var3 follow suit.
    """
    convtype = convtype.lower()
    # Retrieve the appropriate transformation matrix from the constants.
    rgb_matrix = rgb_type.conversion_matrices[convtype]

    logger.debug(
        "  \\* Applying RGB conversion matrix: %s->%s",
        rgb_type.__class__.__name__,
        convtype,
    )
    # Stuff the RGB/XYZ values into a NumPy matrix for conversion.
    var_matrix = numpy.array((var1, var2, var3))
    # Perform the adaptation via matrix multiplication.
    result_matrix = numpy.dot(rgb_matrix, var_matrix)
    rgb_r, rgb_g, rgb_b = result_matrix
    # Clamp these values to a valid range.
    rgb_r = max(rgb_r, 0.0)
    rgb_g = max(rgb_g, 0.0)
    rgb_b = max(rgb_b, 0.0)
    return rgb_r, rgb_g, rgb_b


class ConversionManager(object):
    __metaclass__ = ABCMeta

    def __init__(self):
        self.registered_color_spaces = set()

    def add_type_conversion(self, start_type, target_type, conversion_function):
        """
        Register a conversion function between two color spaces.
        :param start_type: Starting color space.
        :param target_type: Target color space.
        :param conversion_function: Conversion function.
        """
        self.registered_color_spaces.add(start_type)
        self.registered_color_spaces.add(target_type)
        logger.debug("Registered conversion from %s to %s", start_type, target_type)

    @abstractmethod
    def get_conversion_path(self, start_type, target_type):
        """
        Return a list of conversion functions that if applied iteratively on a
        color of the start_type color space result in a color in the result_type
        color space.

        Raises an UndefinedConversionError if no valid conversion path
        can be found.

        :param start_type: Starting color space type.
        :param target_type: Target color space type.
        :return: List of conversion functions.
        """
        pass

    @staticmethod
    def _normalise_type(color_type):
        """
        Return the highest superclass that is valid for color space
        conversions (e.g., AdobeRGB -> BaseRGBColor).
        """
        if issubclass(color_type, BaseRGBColor):
            return BaseRGBColor
        else:
            return color_type


class GraphConversionManager(ConversionManager):
    def __init__(self):
        super(GraphConversionManager, self).__init__()
        self.conversion_graph = networkx.DiGraph()

    def get_conversion_path(self, start_type, target_type):
        start_type = self._normalise_type(start_type)
        target_type = self._normalise_type(target_type)
        try:
            # Retrieve node sequence that leads from start_type to target_type.
            return self._find_shortest_path(start_type, target_type)
        except (networkx.NetworkXNoPath, networkx.NodeNotFound):
            raise UndefinedConversionError(
                start_type, target_type,
            )

    def _find_shortest_path(self, start_type, target_type):
        path = networkx.shortest_path(self.conversion_graph, start_type, target_type)
        # Look up edges between nodes and retrieve the conversion function
        # for each edge.
        return [
            self.conversion_graph.get_edge_data(node_a, node_b)["conversion_function"]
            for node_a, node_b in zip(path[:-1], path[1:])
        ]

    def add_type_conversion(self, start_type, target_type, conversion_function):
        super(GraphConversionManager, self).add_type_conversion(
            start_type, target_type, conversion_function
        )
        self.conversion_graph.add_edge(
            start_type, target_type, conversion_function=conversion_function
        )


class DummyConversionManager(ConversionManager):
    def add_type_conversion(self, start_type, target_type, conversion_function):
        pass

    def get_conversion_path(self, start_type, target_type):
        raise UndefinedConversionError(
            start_type, target_type,
        )


_conversion_manager = GraphConversionManager()


def color_conversion_function(start_type, target_type):
    """
    Decorator to indicate a function that performs a conversion from one color
    space to another.

    This decorator will return the original function unmodified, however it will
    be registered in the _conversion_manager so it can be used to perform color
    space transformations between color spaces that do not have direct
    conversion functions (e.g., Luv to CMYK).

    Note: For a conversion to/from RGB supply the BaseRGBColor class.

    :param start_type: Starting color space type
    :param target_type: Target color space type
    """

    def decorator(f):
        f.start_type = start_type
        f.target_type = target_type
        _conversion_manager.add_type_conversion(start_type, target_type, f)
        return f

    return decorator


# noinspection PyPep8Naming,PyUnusedLocal
@color_conversion_function(SpectralColor, XYZColor)
def Spectral_to_XYZ(cobj, illuminant_override=None, *args, **kwargs):
    """
    Converts spectral readings to XYZ.
    """
    # If the user provides an illuminant_override numpy array, use it.
    if illuminant_override:
        reference_illum = illuminant_override
    else:
        # Otherwise, look up the illuminant from known standards based
        # on the value of 'illuminant' pulled from the SpectralColor object.
        try:
            reference_illum = spectral_constants.REF_ILLUM_TABLE[cobj.illuminant]
        except KeyError:
            raise InvalidIlluminantError(cobj.illuminant)

    # Get the spectral distribution of the selected standard observer.
    if cobj.observer == "10":
        std_obs_x = spectral_constants.STDOBSERV_X10
        std_obs_y = spectral_constants.STDOBSERV_Y10
        std_obs_z = spectral_constants.STDOBSERV_Z10
    else:
        # Assume 2 degree, since it is theoretically the only other possibility.
        std_obs_x = spectral_constants.STDOBSERV_X2
        std_obs_y = spectral_constants.STDOBSERV_Y2
        std_obs_z = spectral_constants.STDOBSERV_Z2

    # This is a NumPy array containing the spectral distribution of the color.
    sample = cobj.get_numpy_array()

    # The denominator is constant throughout the entire calculation for X,
    # Y, and Z coordinates. Calculate it once and re-use.
    denom = std_obs_y * reference_illum

    # This is also a common element in the calculation whereby the sample
    # NumPy array is multiplied by the reference illuminant's power distribution
    # (which is also a NumPy array).
    sample_by_ref_illum = sample * reference_illum

    # Calculate the numerator of the equation to find X.
    x_numerator = sample_by_ref_illum * std_obs_x
    y_numerator = sample_by_ref_illum * std_obs_y
    z_numerator = sample_by_ref_illum * std_obs_z

    xyz_x = x_numerator.sum() / denom.sum()
    xyz_y = y_numerator.sum() / denom.sum()
    xyz_z = z_numerator.sum() / denom.sum()

    return XYZColor(
        xyz_x, xyz_y, xyz_z, observer=cobj.observer, illuminant=cobj.illuminant
    )


# noinspection PyPep8Naming,PyUnusedLocal
@color_conversion_function(LabColor, LCHabColor)
def Lab_to_LCHab(cobj, *args, **kwargs):
    """
    Convert from CIE Lab to LCH(ab).
    """
    lch_l = cobj.lab_l
    lch_c = math.sqrt(math.pow(float(cobj.lab_a), 2) + math.pow(float(cobj.lab_b), 2))
    lch_h = math.atan2(float(cobj.lab_b), float(cobj.lab_a))

    if lch_h > 0:
        lch_h = (lch_h / math.pi) * 180
    else:
        lch_h = 360 - (math.fabs(lch_h) / math.pi) * 180

    return LCHabColor(
        lch_l, lch_c, lch_h, observer=cobj.observer, illuminant=cobj.illuminant
    )


# noinspection PyPep8Naming,PyUnusedLocal
@color_conversion_function(LabColor, XYZColor)
def Lab_to_XYZ(cobj, *args, **kwargs):
    """
    Convert from Lab to XYZ
    """
    illum = cobj.get_illuminant_xyz()
    xyz_y = (cobj.lab_l + 16.0) / 116.0
    xyz_x = cobj.lab_a / 500.0 + xyz_y
    xyz_z = xyz_y - cobj.lab_b / 200.0

    if math.pow(xyz_y, 3) > color_constants.CIE_E:
        xyz_y = math.pow(xyz_y, 3)
    else:
        xyz_y = (xyz_y - 16.0 / 116.0) / 7.787

    if math.pow(xyz_x, 3) > color_constants.CIE_E:
        xyz_x = math.pow(xyz_x, 3)
    else:
        xyz_x = (xyz_x - 16.0 / 116.0) / 7.787

    if math.pow(xyz_z, 3) > color_constants.CIE_E:
        xyz_z = math.pow(xyz_z, 3)
    else:
        xyz_z = (xyz_z - 16.0 / 116.0) / 7.787

    xyz_x = illum["X"] * xyz_x
    xyz_y = illum["Y"] * xyz_y
    xyz_z = illum["Z"] * xyz_z

    return XYZColor(
        xyz_x, xyz_y, xyz_z, observer=cobj.observer, illuminant=cobj.illuminant
    )


# noinspection PyPep8Naming,PyUnusedLocal
@color_conversion_function(LuvColor, LCHuvColor)
def Luv_to_LCHuv(cobj, *args, **kwargs):
    """
    Convert from CIE Luv to LCH(uv).
    """
    lch_l = cobj.luv_l
    lch_c = math.sqrt(math.pow(cobj.luv_u, 2.0) + math.pow(cobj.luv_v, 2.0))
    lch_h = math.atan2(float(cobj.luv_v), float(cobj.luv_u))

    if lch_h > 0:
        lch_h = (lch_h / math.pi) * 180
    else:
        lch_h = 360 - (math.fabs(lch_h) / math.pi) * 180
    return LCHuvColor(
        lch_l, lch_c, lch_h, observer=cobj.observer, illuminant=cobj.illuminant
    )


# noinspection PyPep8Naming,PyUnusedLocal
@color_conversion_function(LuvColor, XYZColor)
def Luv_to_XYZ(cobj, *args, **kwargs):
    """
    Convert from Luv to XYZ.
    """
    illum = cobj.get_illuminant_xyz()
    # Without Light, there is no color. Short-circuit this and avoid some
    # zero division errors in the var_a_frac calculation.
    if cobj.luv_l <= 0.0:
        xyz_x = 0.0
        xyz_y = 0.0
        xyz_z = 0.0
        return XYZColor(
            xyz_x, xyz_y, xyz_z, observer=cobj.observer, illuminant=cobj.illuminant
        )

    # Various variables used throughout the conversion.
    cie_k_times_e = color_constants.CIE_K * color_constants.CIE_E
    u_sub_0 = (4.0 * illum["X"]) / (illum["X"] + 15.0 * illum["Y"] + 3.0 * illum["Z"])
    v_sub_0 = (9.0 * illum["Y"]) / (illum["X"] + 15.0 * illum["Y"] + 3.0 * illum["Z"])
    var_u = cobj.luv_u / (13.0 * cobj.luv_l) + u_sub_0
    var_v = cobj.luv_v / (13.0 * cobj.luv_l) + v_sub_0

    # Y-coordinate calculations.
    if cobj.luv_l > cie_k_times_e:
        xyz_y = math.pow((cobj.luv_l + 16.0) / 116.0, 3.0)
    else:
        xyz_y = cobj.luv_l / color_constants.CIE_K

    # X-coordinate calculation.
    xyz_x = xyz_y * 9.0 * var_u / (4.0 * var_v)
    # Z-coordinate calculation.
    xyz_z = xyz_y * (12.0 - 3.0 * var_u - 20.0 * var_v) / (4.0 * var_v)

    return XYZColor(
        xyz_x, xyz_y, xyz_z, illuminant=cobj.illuminant, observer=cobj.observer
    )


# noinspection PyPep8Naming,PyUnusedLocal
@color_conversion_function(LCHabColor, LabColor)
def LCHab_to_Lab(cobj, *args, **kwargs):
    """
    Convert from LCH(ab) to Lab.
    """
    lab_l = cobj.lch_l
    lab_a = math.cos(math.radians(cobj.lch_h)) * cobj.lch_c
    lab_b = math.sin(math.radians(cobj.lch_h)) * cobj.lch_c
    return LabColor(
        lab_l, lab_a, lab_b, illuminant=cobj.illuminant, observer=cobj.observer
    )


# noinspection PyPep8Naming,PyUnusedLocal
@color_conversion_function(LCHuvColor, LuvColor)
def LCHuv_to_Luv(cobj, *args, **kwargs):
    """
    Convert from LCH(uv) to Luv.
    """
    luv_l = cobj.lch_l
    luv_u = math.cos(math.radians(cobj.lch_h)) * cobj.lch_c
    luv_v = math.sin(math.radians(cobj.lch_h)) * cobj.lch_c
    return LuvColor(
        luv_l, luv_u, luv_v, illuminant=cobj.illuminant, observer=cobj.observer
    )


# noinspection PyPep8Naming,PyUnusedLocal
@color_conversion_function(xyYColor, XYZColor)
def xyY_to_XYZ(cobj, *args, **kwargs):
    """
    Convert from xyY to XYZ.
    """
    # avoid division by zero
    if cobj.xyy_y == 0.0:
        xyz_x = 0.0
        xyz_y = 0.0
        xyz_z = 0.0
    else:
        xyz_x = (cobj.xyy_x * cobj.xyy_Y) / cobj.xyy_y
        xyz_y = cobj.xyy_Y
        xyz_z = ((1.0 - cobj.xyy_x - cobj.xyy_y) * xyz_y) / cobj.xyy_y

    return XYZColor(
        xyz_x, xyz_y, xyz_z, illuminant=cobj.illuminant, observer=cobj.observer
    )


# noinspection PyPep8Naming,PyUnusedLocal
@color_conversion_function(XYZColor, xyYColor)
def XYZ_to_xyY(cobj, *args, **kwargs):
    """
    Convert from XYZ to xyY.
    """
    xyz_sum = cobj.xyz_x + cobj.xyz_y + cobj.xyz_z
    # avoid division by zero
    if xyz_sum == 0.0:
        xyy_x = 0.0
        xyy_y = 0.0
    else:
        xyy_x = cobj.xyz_x / xyz_sum
        xyy_y = cobj.xyz_y / xyz_sum
    xyy_Y = cobj.xyz_y

    return xyYColor(
        xyy_x, xyy_y, xyy_Y, observer=cobj.observer, illuminant=cobj.illuminant
    )


# noinspection PyPep8Naming,PyUnusedLocal
@color_conversion_function(XYZColor, LuvColor)
def XYZ_to_Luv(cobj, *args, **kwargs):
    """
    Convert from XYZ to Luv
    """
    temp_x = cobj.xyz_x
    temp_y = cobj.xyz_y
    temp_z = cobj.xyz_z
    denom = temp_x + (15.0 * temp_y) + (3.0 * temp_z)
    # avoid division by zero
    if denom == 0.0:
        luv_u = 0.0
        luv_v = 0.0
    else:
        luv_u = (4.0 * temp_x) / denom
        luv_v = (9.0 * temp_y) / denom

    illum = cobj.get_illuminant_xyz()
    temp_y = temp_y / illum["Y"]
    if temp_y > color_constants.CIE_E:
        temp_y = math.pow(temp_y, (1.0 / 3.0))
    else:
        temp_y = (7.787 * temp_y) + (16.0 / 116.0)

    ref_U = (4.0 * illum["X"]) / (illum["X"] + (15.0 * illum["Y"]) + (3.0 * illum["Z"]))
    ref_V = (9.0 * illum["Y"]) / (illum["X"] + (15.0 * illum["Y"]) + (3.0 * illum["Z"]))

    luv_l = (116.0 * temp_y) - 16.0
    luv_u = 13.0 * luv_l * (luv_u - ref_U)
    luv_v = 13.0 * luv_l * (luv_v - ref_V)

    return LuvColor(
        luv_l, luv_u, luv_v, observer=cobj.observer, illuminant=cobj.illuminant
    )


# noinspection PyPep8Naming,PyUnusedLocal
@color_conversion_function(XYZColor, LabColor)
def XYZ_to_Lab(cobj, *args, **kwargs):
    """
    Converts XYZ to Lab.
    """
    illum = cobj.get_illuminant_xyz()
    temp_x = cobj.xyz_x / illum["X"]
    temp_y = cobj.xyz_y / illum["Y"]
    temp_z = cobj.xyz_z / illum["Z"]

    if temp_x > color_constants.CIE_E:
        temp_x = math.pow(temp_x, (1.0 / 3.0))
    else:
        temp_x = (7.787 * temp_x) + (16.0 / 116.0)

    if temp_y > color_constants.CIE_E:
        temp_y = math.pow(temp_y, (1.0 / 3.0))
    else:
        temp_y = (7.787 * temp_y) + (16.0 / 116.0)

    if temp_z > color_constants.CIE_E:
        temp_z = math.pow(temp_z, (1.0 / 3.0))
    else:
        temp_z = (7.787 * temp_z) + (16.0 / 116.0)

    lab_l = (116.0 * temp_y) - 16.0
    lab_a = 500.0 * (temp_x - temp_y)
    lab_b = 200.0 * (temp_y - temp_z)
    return LabColor(
        lab_l, lab_a, lab_b, observer=cobj.observer, illuminant=cobj.illuminant
    )


# noinspection PyPep8Naming,PyUnusedLocal
@color_conversion_function(XYZColor, BaseRGBColor)
def XYZ_to_RGB(cobj, target_rgb, *args, **kwargs):
    """
    XYZ to RGB conversion.
    """
    temp_X = cobj.xyz_x
    temp_Y = cobj.xyz_y
    temp_Z = cobj.xyz_z

    logger.debug("  \\- Target RGB space: %s", target_rgb)
    target_illum = target_rgb.native_illuminant
    logger.debug("  \\- Target native illuminant: %s", target_illum)
    logger.debug("  \\- XYZ color's illuminant: %s", cobj.illuminant)

    # If the XYZ values were taken with a different reference white than the
    # native reference white of the target RGB space, a transformation matrix
    # must be applied.
    if cobj.illuminant != target_illum:
        logger.debug(
            "  \\* Applying transformation from %s to %s ",
            cobj.illuminant,
            target_illum,
        )
        # Get the adjusted XYZ values, adapted for the target illuminant.
        temp_X, temp_Y, temp_Z = apply_chromatic_adaptation(
            temp_X, temp_Y, temp_Z, orig_illum=cobj.illuminant, targ_illum=target_illum
        )
        logger.debug("  \\*   New values: %.3f, %.3f, %.3f", temp_X, temp_Y, temp_Z)

    # Apply an RGB working space matrix to the XYZ values (matrix mul).
    rgb_r, rgb_g, rgb_b = apply_RGB_matrix(
        temp_X, temp_Y, temp_Z, rgb_type=target_rgb, convtype="xyz_to_rgb"
    )

    # v
    linear_channels = dict(r=rgb_r, g=rgb_g, b=rgb_b)
    # V
    nonlinear_channels = {}
    if target_rgb == sRGBColor:
        for channel in ["r", "g", "b"]:
            v = linear_channels[channel]
            if v <= 0.0031308:
                nonlinear_channels[channel] = v * 12.92
            else:
                nonlinear_channels[channel] = 1.055 * math.pow(v, 1 / 2.4) - 0.055
    elif target_rgb == BT2020Color:
        if kwargs.get("is_12_bits_system"):
            a, b = 1.0993, 0.0181
        else:
            a, b = 1.099, 0.018
        for channel in ["r", "g", "b"]:
            v = linear_channels[channel]
            if v < b:
                nonlinear_channels[channel] = v * 4.5
            else:
                nonlinear_channels[channel] = a * math.pow(v, 0.45) - (a - 1)
    else:
        # If it's not sRGB...
        for channel in ["r", "g", "b"]:
            v = linear_channels[channel]
            nonlinear_channels[channel] = math.pow(v, 1 / target_rgb.rgb_gamma)

    return target_rgb(
        nonlinear_channels["r"], nonlinear_channels["g"], nonlinear_channels["b"]
    )


# noinspection PyPep8Naming,PyUnusedLocal
@color_conversion_function(BaseRGBColor, XYZColor)
def RGB_to_XYZ(cobj, target_illuminant=None, *args, **kwargs):
    """
    RGB to XYZ conversion. Expects RGB values between 0 and 255.

    Based off of: http://www.brucelindbloom.com/index.html?Eqn_RGB_to_XYZ.html
    """
    # Will contain linearized RGB channels (removed the gamma func).
    linear_channels = {}

    if isinstance(cobj, sRGBColor):
        for channel in ["r", "g", "b"]:
            V = getattr(cobj, "rgb_" + channel)
            if V <= 0.04045:
                linear_channels[channel] = V / 12.92
            else:
                linear_channels[channel] = math.pow((V + 0.055) / 1.055, 2.4)
    elif isinstance(cobj, BT2020Color):
        if kwargs.get("is_12_bits_system"):
            a, b, c = 1.0993, 0.0181, 0.081697877417347  # noqa
        else:
            a, b, c = 1.099, 0.018, 0.08124794403514049  # noqa
        for channel in ["r", "g", "b"]:
            V = getattr(cobj, "rgb_" + channel)
            if V <= c:
                linear_channels[channel] = V / 4.5
            else:
                linear_channels[channel] = math.pow((V + (a - 1)) / a, 1 / 0.45)
    else:
        # If it's not sRGB...
        gamma = cobj.rgb_gamma

        for channel in ["r", "g", "b"]:
            V = getattr(cobj, "rgb_" + channel)
            linear_channels[channel] = math.pow(V, gamma)

    # Apply an RGB working space matrix to the XYZ values (matrix mul).
    xyz_x, xyz_y, xyz_z = apply_RGB_matrix(
        linear_channels["r"],
        linear_channels["g"],
        linear_channels["b"],
        rgb_type=cobj,
        convtype="rgb_to_xyz",
    )

    if target_illuminant is None:
        target_illuminant = cobj.native_illuminant

    # The illuminant of the original RGB object. This will always match
    # the RGB colorspace's native illuminant.
    illuminant = cobj.native_illuminant
    xyzcolor = XYZColor(xyz_x, xyz_y, xyz_z, illuminant=illuminant)
    # This will take care of any illuminant changes for us (if source
    # illuminant != target illuminant).
    xyzcolor.apply_adaptation(target_illuminant)

    return xyzcolor


# noinspection PyPep8Naming,PyUnusedLocal
def __RGB_to_Hue(var_R, var_G, var_B, var_min, var_max):
    """
    For RGB_to_HSL and RGB_to_HSV, the Hue (H) component is calculated in
    the same way.
    """
    if var_max == var_min:
        return 0.0
    elif var_max == var_R:
        return (60.0 * ((var_G - var_B) / (var_max - var_min)) + 360) % 360.0
    elif var_max == var_G:
        return 60.0 * ((var_B - var_R) / (var_max - var_min)) + 120
    elif var_max == var_B:
        return 60.0 * ((var_R - var_G) / (var_max - var_min)) + 240.0


# noinspection PyPep8Naming,PyUnusedLocal
@color_conversion_function(BaseRGBColor, HSVColor)
def RGB_to_HSV(cobj, *args, **kwargs):
    """
    Converts from RGB to HSV.

    H values are in degrees and are 0 to 360.
    S values are a percentage, 0.0 to 1.0.
    V values are a percentage, 0.0 to 1.0.
    """
    var_R = cobj.rgb_r
    var_G = cobj.rgb_g
    var_B = cobj.rgb_b

    var_max = max(var_R, var_G, var_B)
    var_min = min(var_R, var_G, var_B)

    var_H = __RGB_to_Hue(var_R, var_G, var_B, var_min, var_max)

    if var_max == 0:
        var_S = 0
    else:
        var_S = 1.0 - (var_min / var_max)

    var_V = var_max

    return HSVColor(var_H, var_S, var_V)


# noinspection PyPep8Naming,PyUnusedLocal
@color_conversion_function(BaseRGBColor, HSLColor)
def RGB_to_HSL(cobj, *args, **kwargs):
    """
    Converts from RGB to HSL.

    H values are in degrees and are 0 to 360.
    S values are a percentage, 0.0 to 1.0.
    L values are a percentage, 0.0 to 1.0.
    """
    var_R = cobj.rgb_r
    var_G = cobj.rgb_g
    var_B = cobj.rgb_b

    var_max = max(var_R, var_G, var_B)
    var_min = min(var_R, var_G, var_B)

    var_H = __RGB_to_Hue(var_R, var_G, var_B, var_min, var_max)
    var_L = 0.5 * (var_max + var_min)

    if var_max == var_min:
        var_S = 0
    elif var_L <= 0.5:
        var_S = (var_max - var_min) / (2.0 * var_L)
    else:
        var_S = (var_max - var_min) / (2.0 - (2.0 * var_L))

    return HSLColor(var_H, var_S, var_L)


# noinspection PyPep8Naming,PyUnusedLocal
def __Calc_HSL_to_RGB_Components(var_q, var_p, C):
    """
    This is used in HSL_to_RGB conversions on R, G, and B.
    """
    if C < 0:
        C += 1.0
    if C > 1:
        C -= 1.0

    # Computing C of vector (Color R, Color G, Color B)
    if C < (1.0 / 6.0):
        return var_p + ((var_q - var_p) * 6.0 * C)
    elif (1.0 / 6.0) <= C < 0.5:
        return var_q
    elif 0.5 <= C < (2.0 / 3.0):
        return var_p + ((var_q - var_p) * 6.0 * ((2.0 / 3.0) - C))
    else:
        return var_p


# noinspection PyPep8Naming,PyUnusedLocal
@color_conversion_function(HSVColor, BaseRGBColor)
def HSV_to_RGB(cobj, target_rgb, *args, **kwargs):
    """
    HSV to RGB conversion.

    H values are in degrees and are 0 to 360.
    S values are a percentage, 0.0 to 1.0.
    V values are a percentage, 0.0 to 1.0.
    """
    H = cobj.hsv_h
    S = cobj.hsv_s
    V = cobj.hsv_v

    h_floored = int(math.floor(H))
    h_sub_i = int(h_floored / 60) % 6
    var_f = (H / 60.0) - (h_floored // 60)
    var_p = V * (1.0 - S)
    var_q = V * (1.0 - var_f * S)
    var_t = V * (1.0 - (1.0 - var_f) * S)

    if h_sub_i == 0:
        rgb_r = V
        rgb_g = var_t
        rgb_b = var_p
    elif h_sub_i == 1:
        rgb_r = var_q
        rgb_g = V
        rgb_b = var_p
    elif h_sub_i == 2:
        rgb_r = var_p
        rgb_g = V
        rgb_b = var_t
    elif h_sub_i == 3:
        rgb_r = var_p
        rgb_g = var_q
        rgb_b = V
    elif h_sub_i == 4:
        rgb_r = var_t
        rgb_g = var_p
        rgb_b = V
    elif h_sub_i == 5:
        rgb_r = V
        rgb_g = var_p
        rgb_b = var_q
    else:
        raise ValueError("Unable to convert HSL->RGB due to value error.")

    # TODO: Investigate intent of following code block.
    # In the event that they define an HSV color and want to convert it to
    # a particular RGB space, let them override it here.
    # if target_rgb is not None:
    #     rgb_type = target_rgb
    # else:
    #     rgb_type = cobj.rgb_type

    return target_rgb(rgb_r, rgb_g, rgb_b)


# noinspection PyPep8Naming,PyUnusedLocal
@color_conversion_function(HSLColor, BaseRGBColor)
def HSL_to_RGB(cobj, target_rgb, *args, **kwargs):
    """
    HSL to RGB conversion.
    """
    H = cobj.hsl_h
    S = cobj.hsl_s
    L = cobj.hsl_l

    if L < 0.5:
        var_q = L * (1.0 + S)
    else:
        var_q = L + S - (L * S)

    var_p = 2.0 * L - var_q

    # H normalized to range [0,1]
    h_sub_k = H / 360.0

    t_sub_R = h_sub_k + (1.0 / 3.0)
    t_sub_G = h_sub_k
    t_sub_B = h_sub_k - (1.0 / 3.0)

    rgb_r = __Calc_HSL_to_RGB_Components(var_q, var_p, t_sub_R)
    rgb_g = __Calc_HSL_to_RGB_Components(var_q, var_p, t_sub_G)
    rgb_b = __Calc_HSL_to_RGB_Components(var_q, var_p, t_sub_B)

    # TODO: Investigate intent of following code block.
    # In the event that they define an HSV color and want to convert it to
    # a particular RGB space, let them override it here.
    # if target_rgb is not None:
    #     rgb_type = target_rgb
    # else:
    #     rgb_type = cobj.rgb_type

    return target_rgb(rgb_r, rgb_g, rgb_b)


# noinspection PyPep8Naming,PyUnusedLocal
@color_conversion_function(BaseRGBColor, CMYColor)
def RGB_to_CMY(cobj, *args, **kwargs):
    """
    RGB to CMY conversion.

    NOTE: CMYK and CMY values range from 0.0 to 1.0
    """
    cmy_c = 1.0 - cobj.rgb_r
    cmy_m = 1.0 - cobj.rgb_g
    cmy_y = 1.0 - cobj.rgb_b

    return CMYColor(cmy_c, cmy_m, cmy_y)


# noinspection PyPep8Naming,PyUnusedLocal
@color_conversion_function(CMYColor, BaseRGBColor)
def CMY_to_RGB(cobj, target_rgb, *args, **kwargs):
    """
    Converts CMY to RGB via simple subtraction.

    NOTE: Returned values are in the range of 0-255.
    """
    rgb_r = 1.0 - cobj.cmy_c
    rgb_g = 1.0 - cobj.cmy_m
    rgb_b = 1.0 - cobj.cmy_y

    return target_rgb(rgb_r, rgb_g, rgb_b)


# noinspection PyPep8Naming,PyUnusedLocal
@color_conversion_function(CMYColor, CMYKColor)
def CMY_to_CMYK(cobj, *args, **kwargs):
    """
    Converts from CMY to CMYK.

    NOTE: CMYK and CMY values range from 0.0 to 1.0
    """
    var_k = 1.0
    if cobj.cmy_c < var_k:
        var_k = cobj.cmy_c
    if cobj.cmy_m < var_k:
        var_k = cobj.cmy_m
    if cobj.cmy_y < var_k:
        var_k = cobj.cmy_y

    if var_k == 1:
        cmyk_c = 0.0
        cmyk_m = 0.0
        cmyk_y = 0.0
    else:
        cmyk_c = (cobj.cmy_c - var_k) / (1.0 - var_k)
        cmyk_m = (cobj.cmy_m - var_k) / (1.0 - var_k)
        cmyk_y = (cobj.cmy_y - var_k) / (1.0 - var_k)
    cmyk_k = var_k

    return CMYKColor(cmyk_c, cmyk_m, cmyk_y, cmyk_k)


# noinspection PyPep8Naming,PyUnusedLocal
@color_conversion_function(CMYKColor, CMYColor)
def CMYK_to_CMY(cobj, *args, **kwargs):
    """
    Converts CMYK to CMY.

    NOTE: CMYK and CMY values range from 0.0 to 1.0
    """
    cmy_c = cobj.cmyk_c * (1.0 - cobj.cmyk_k) + cobj.cmyk_k
    cmy_m = cobj.cmyk_m * (1.0 - cobj.cmyk_k) + cobj.cmyk_k
    cmy_y = cobj.cmyk_y * (1.0 - cobj.cmyk_k) + cobj.cmyk_k

    return CMYColor(cmy_c, cmy_m, cmy_y)


# noinspection PyPep8Naming,PyUnusedLocal
@color_conversion_function(XYZColor, IPTColor)
def XYZ_to_IPT(cobj, *args, **kwargs):
    """
    Converts XYZ to IPT.

    NOTE: XYZ values need to be adapted to 2 degree D65

    Reference:
    Fairchild, M. D. (2013). Color appearance models, 3rd Ed. (pp. 271-272).
    John Wiley & Sons.
    """
    if cobj.illuminant != "d65" or cobj.observer != "2":
        raise ValueError("XYZColor for XYZ->IPT conversion needs to be D65 adapted.")
    xyz_values = numpy.array(cobj.get_value_tuple())
    lms_values = numpy.dot(IPTColor.conversion_matrices["xyz_to_lms"], xyz_values)

    lms_prime = numpy.sign(lms_values) * numpy.abs(lms_values) ** 0.43

    ipt_values = numpy.dot(IPTColor.conversion_matrices["lms_to_ipt"], lms_prime)
    return IPTColor(*ipt_values)


# noinspection PyPep8Naming,PyUnusedLocal
@color_conversion_function(IPTColor, XYZColor)
def IPT_to_XYZ(cobj, *args, **kwargs):
    """
    Converts IPT to XYZ.
    """
    ipt_values = numpy.array(cobj.get_value_tuple())
    lms_values = numpy.dot(
        numpy.linalg.inv(IPTColor.conversion_matrices["lms_to_ipt"]), ipt_values
    )

    lms_prime = numpy.sign(lms_values) * numpy.abs(lms_values) ** (1 / 0.43)

    xyz_values = numpy.dot(
        numpy.linalg.inv(IPTColor.conversion_matrices["xyz_to_lms"]), lms_prime
    )
    return XYZColor(*xyz_values, observer="2", illuminant="d65")


# We use this as a template conversion dict for each RGB color space. They
# are all identical.
_RGB_CONVERSION_DICT_TEMPLATE = {
    "HSLColor": [RGB_to_HSL],
    "HSVColor": [RGB_to_HSV],
    "CMYColor": [RGB_to_CMY],
    "CMYKColor": [RGB_to_CMY, CMY_to_CMYK],
    "XYZColor": [RGB_to_XYZ],
    "xyYColor": [RGB_to_XYZ, XYZ_to_xyY],
    "LabColor": [RGB_to_XYZ, XYZ_to_Lab],
    "LCHabColor": [RGB_to_XYZ, XYZ_to_Lab, Lab_to_LCHab],
    "LCHuvColor": [RGB_to_XYZ, XYZ_to_Luv, Luv_to_LCHuv],
    "LuvColor": [RGB_to_XYZ, XYZ_to_Luv],
    "IPTColor": [RGB_to_XYZ, XYZ_to_IPT],
}


def convert_color(
    color,
    target_cs,
    through_rgb_type=sRGBColor,
    target_illuminant=None,
    *args,
    **kwargs
):
    """
    Converts the color to the designated color space.

    :param color: A Color instance to convert.
    :param target_cs: The Color class to convert to. Note that this is not
        an instance, but a class.
    :keyword BaseRGBColor through_rgb_type: If during your conversion between
        your original and target color spaces you have to pass through RGB,
        this determines which kind of RGB to use. For example, XYZ->HSL.
        You probably don't need to specify this unless you have a special
        usage case.
    :type target_illuminant: None or str
    :keyword target_illuminant: If during conversion from RGB to a reflective
        color space you want to explicitly end up with a certain illuminant,
        pass this here. Otherwise the RGB space's native illuminant
        will be used.
    :returns: An instance of the type passed in as ``target_cs``.
    :raises: :py:exc:`colormath.color_exceptions.UndefinedConversionError`
        if conversion between the two color spaces isn't possible.
    """
    if isinstance(target_cs, str):
        raise ValueError("target_cs parameter must be a Color object.")
    if not issubclass(target_cs, ColorBase):
        raise ValueError("target_cs parameter must be a Color object.")

    conversions = _conversion_manager.get_conversion_path(color.__class__, target_cs)

    logger.debug("Converting %s to %s", color, target_cs)
    logger.debug(" @ Conversion path: %s", conversions)

    # Start with original color in case we convert to the same color space.
    new_color = color

    if issubclass(target_cs, BaseRGBColor):
        # If the target_cs is an RGB color space of some sort, then we
        # have to set our through_rgb_type to make sure the conversion returns
        # the expected RGB colorspace (instead of defaulting to sRGBColor).
        through_rgb_type = target_cs

    # We have to be careful to use the same RGB color space that created
    # an object (if it was created by a conversion) in order to get correct
    # results. For example, XYZ->HSL via Adobe RGB should default to Adobe
    # RGB when taking that generated HSL object back to XYZ.
    # noinspection PyProtectedMember
    if through_rgb_type != sRGBColor:
        # User overrides take priority over everything.
        # noinspection PyProtectedMember
        target_rgb = through_rgb_type
    elif color._through_rgb_type:
        # Otherwise, a value on the color object is the next best thing,
        # when available.
        # noinspection PyProtectedMember
        target_rgb = color._through_rgb_type
    else:
        # We could collapse this into a single if statement above,
        # but I think this reads better.
        target_rgb = through_rgb_type

    # Iterate through the list of functions for the conversion path, storing
    # the results in a dictionary via update(). This way the user has access
    # to all of the variables involved in the conversion.
    for func in conversions:
        # Execute the function in this conversion step and store the resulting
        # Color object.
        logger.debug(
            " * Conversion: %s passed to %s()", new_color.__class__.__name__, func
        )
        logger.debug(" |->  in %s", new_color)

        if func:
            # This can be None if you try to convert a color to the color
            # space that is already in. IE: XYZ->XYZ.
            new_color = func(
                new_color,
                target_rgb=target_rgb,
                target_illuminant=target_illuminant,
                *args,
                **kwargs
            )

        logger.debug(" |-< out %s", new_color)

    # If this conversion had something other than the default sRGB color space
    # requested,
    if through_rgb_type != sRGBColor:
        new_color._through_rgb_type = through_rgb_type

    return new_color
